# -*- coding: iso-8859-1 -*-

# ==============================================================================
# COPYRIGHT (C) 1991 - 2003  EDF R&D                  WWW.CODE-ASTER.ORG
# THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
# IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
# THE FREE SOFTWARE FOUNDATION; EITHER VERSION 2 OF THE LICENSE, OR
# (AT YOUR OPTION) ANY LATER VERSION.
#
# THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
# WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. SEE THE GNU
# GENERAL PUBLIC LICENSE FOR MORE DETAILS.
#
# YOU SHOULD HAVE RECEIVED A COPY OF THE GNU GENERAL PUBLIC LICENSE
# ALONG WITH THIS PROGRAM; IF NOT, WRITE TO EDF R&D CODE_ASTER,
#    1 AVENUE DU GENERAL DE GAULLE, 92141 CLAMART CEDEX, FRANCE.
# ==============================================================================

"""Definition of ITEM class representing a MySQL item.
"""

__revision__ = "$Id: schema.py 2626 2006-12-18 07:50:27Z courtois $"


# ----- differ messages translation
def _(mesg):
   return mesg

import os
import time
import sha
import datetime
import re
from types import StringTypes, DictType, IntType, LongType

from myconnect import MySQLError
from mystring  import convert

__verbose__ = False
__all__ = ['ITEM', 'ReadDB', 'USER', 'INTERVENANT', 'MSG', 'PRODUIT', 'VERSION',
           'PROJET', 'STATUS', 'TYPE', 'ISSUE',
           'PRODUIT_VERSION', 'ISSUE_MESSAGES', 'YES', 'NO',
           'ReadDBError', 'WriteDBError', 'ConnectDBError']

mysql_date_fmt = '%Y-%m-%d %H:%M:%S'
now = time.strftime(mysql_date_fmt, time.gmtime())

# boolean values
YES = 1
NO  = 0

# ------------------------------------------------------------------------------
def transpose(liste):
   """Converts a such list ((x1, x2, ..., xn), (y1, y2, ..., yn), ...)
                      into ((x1, y1, ...), (x2, y2, ...), ..., (xn, yn, ...))
   """
   n = range(len(liste[0]))
   m = range(len(liste))
   liste_t = [[] for i in n]
   for i in n :
      for j in m :
         liste_t[i].append(liste[j][i])
   return liste_t

fmt_val = '%-20s = %s'
fmt_vr  = '%-20s = %s (@)'
fmt_id  = "<--- item référencé dans '%s' avec id = %s"
fmt_fin = "     --->"
fmt_i   = '%8d'
fmt_date = '%d/%m/%Y'
# regexp associée à mysql_date_fmt
re_date  = '([0-9]{4}\-[0-9]{2}\-[0-9]{2}) [0-9]{2}:[0-9]{2}:[0-9]{2}'

# ------------------------------------------------------------------------------
class ConnectDBError(Exception):
   """Error with MySQL database connection."""
   pass

class ReadDBError(Exception):
   """Error when reading items from database."""
   pass

class WriteDBError(Exception):
   """Error when writing items to database."""
   pass

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
class ITEM:
   """Enregistrement d'une table avec éventuellement des références à
   des enregistrements d'autres tables.
   (Je pense que hyperdb de Roundup pourrait être utilisée, mais je n'ai pas
   trouvé comment...!).
   Attributs :
      table    : nom de la table associée à ce type d'enregistrement
      c        : connexion MySQL
      refs     : dictionnaire indiquant la sous-classe des ITEMs référencés
      idkey    : vaut 'id' sauf pour les tables multi-link où il vaut 'linkid'
      primkeys : liste des clés "primaires" (permettant de retrouver un item)
      values   : dictionnaire des champs
      default  : contenu d'item par défaut
      inDB     : True si on a vérifié que l'enregistrement est dans la base
      lu       : True si on a déjà lu l'enregistrement dans la base
      cols     : (cache) ordre des colonnes dans la table (pour SELECT * par ex)
      
      link_class : table où sont stockés les "multi-link"
      parent_class : la classe "parent"

   A priori, tant que lu=False, on n'est pas assuré que les champs qui font
   référence à un autre enregistrement contienne l'ITEM référencé, ils peuvent
   contenir que l'id.
   """
   table = None
   link_class = None
   parent_class = None
   default = {}
   refs = {}
   primkeys = ['id']
   idkey = 'id'
   std_encoding = 'iso-8859-1'
   db_encoding  = 'utf-8'
   # this is used for Roundup schema where max ids are store in 'ids' table.
   tab_idmax = 'ids'

   def __init__(self, dini=None, c=None):
      """Initialisation avec dini qui est :
         - soit le dictionnaire des valeurs des champs de l'enregistrement,
         - soit la valeur du champ "primaire".
      Sont équivalents (en supposant que 'id' est la clé primaire):
         ITEM({'id' : num}, c) et ITEM(num, c)
      """
      toread = False
      if dini == None:
         dini = {}
      elif not type(dini) is DictType:
         dini = { self.idkey : dini }
         toread = True
      self.c      = c
      self.values = self.default.copy()
      self.values.update(dini)
      self.inDB   = False
      self.lu     = False
      self.cols   = None
      if toread:
         self.read()

   def __setitem__(self, key, value):
      """Positionne un champ de l'enregistrement"""
      self.values[key] = value

   def __getitem__(self, key):
      """Retourne la valeur d'un champ de l'enregistrement"""
      return self.values[key]

   def __setid(self, value):
      """Positionne le champ idkey de l'objet"""
      if not self.idkey in self.refs:
         self.values[self.idkey] = value
      elif __verbose__:
         print 'Warning <setid>: %s is an ITEM object.' % repr(self.idkey)

   def __getid(self):
      """Retourne la valeur du champ idkey de l'objet"""
      value = self.values[self.idkey]
      if not self.idkey in self.refs or not isinstance(value, ITEM):
         return value
      else:
         if __verbose__:
            print 'Warning <getid>: %s is an ITEM object.' % repr(self.idkey)
         return value.get(value.idkey)

   def get(self, key, default=None):
      """Retourne la valeur d'un champ de l'enregistrement ou default"""
      return self.values.get(key, default)

   def getrepr(self, refs=True):
      """Affichage. Si 'refs'=False, on n'imprime pas les sous-items."""
      txt = []
      for k, v in self.values.items():
         if k in self.refs.keys():
            item = None
            if isinstance(self[k], ITEM):
               item = v
               idref = v.get(v.idkey)
            else:
               item = eval('%s(%s)' % (self.refs[k], 'c=self.c'))
               try:
                  item.read({item.idkey:v})
                  idref = v
               except (ConnectDBError, ReadDBError), msg:
                  item = None
                  if __verbose__:
                     print msg
            if item != None and (item.__class__.__name__ != self.__class__.__name__\
                               or idref != self.get(self.idkey)):
               if not refs:
                  txt.append(fmt_vr % (k, idref))
               else:
                  txt.append(fmt_id % (k, idref))
                  txt.append(os.linesep.join(['      > %s' \
                        % l for l in repr(item).split(os.linesep)]))
                  txt.append(fmt_fin)
            else:
               txt.append(fmt_vr % (k, v))
         else:
            txt.append(fmt_val % (k, v))
      return os.linesep.join(txt)

   def __repr__(self):
      """Affichage"""
      return self.getrepr()

   def repr(self, refs=False):
      """Affichage"""
      print self.getrepr(refs)

   def prim_crit(self):
      """Critères suffisants pour retrouver l'id d'un enregistrement,
      construit à partir des primkeys."""
      d = {}
      for k in self.primkeys:
         if isinstance(self[k], ITEM):
            d[k] = self[k].get(self[k].idkey)
         else:
            d[k] = self[k]
      return d

   def GetPrimValue(self):
      """Return the value of `primkeys` field.
      Return None if there was more than one primary keys.
      """
      if len(self.primkeys) != 1:
         value = None
      else:
         value = self.get(self.primkeys[0], None)
      return value
   
   def isinDB(self, dcrit=None):
      """Retourne True si l'enregistrement est présent dans la base."""
      if not self.inDB:
         try:
            idv = self.__getid_from_db(dcrit)
            self.inDB = True
         except ReadDBError:
            pass
      return self.inDB

   def read(self, dcrit=None, force=False):
      """Lit l'enregistrement dans la base à partir de prim_crit() ou du
      critère fourni sous forme de dict, il faut donc au moins initialiser les
      champs qui servent au critère (sinon KeyError).
      Si force=True, on ne tient pas compte de l'attrbut lu."""
      if self.c == None:
         raise ConnectDBError, _('no database connection')
      elif force or not self.lu:
         self.__cachecols()
         if dcrit == None:
            dcrit = self.prim_crit()
         criteres = ' AND '.join([
            "%s='%s'" % (k, v) for k, v in dcrit.items() ])
         query = """SELECT * FROM %s WHERE %s;""" % (self.table, criteres)
         res = self.c.exe(query)
         if len(res) != 1:
            raise ReadDBError, _('no result for """%s"""') % query
         else:
            self.lu = True
            self.inDB = True
            lkv = transpose([self.cols, res[0]])
            # effacer le contenu si force=True
            if force:
               self.values = self.default.copy()
            # il faut lire l'id des références avant de les traiter
            for k, v in lkv:
               if isinstance(v, datetime.datetime) or isinstance(v, datetime.date) \
                     or type(v).__name__ == 'DateTime':
                  v = v.strftime(mysql_date_fmt)
               self[k] = v
            self.__decode_string_fromDB()
            # traitement des références
            for k, v in [[k, v] for k, v in lkv if k in self.refs.keys()]:
               if self.refs[k] == self.__class__.__name__ and self.__getid() == v:
                  #print "Warning: %s avec id=%s s'auto-référence." \
                  #   % (self.refs[k], v)
                  pass
               # s'il n'y a pas de référence
               elif v != None:
                  self[k] = eval('%s(%s)' % (self.refs[k], 'c=self.c'))
                  self[k].read({self[k].idkey:v})

   def write(self, force=False, force_recursive=False):
      """Ecrit l'enregistrement dans la base."""
      if force_recursive:
         force = True
      if self.c == None:
         raise ConnectDBError, _('no database connection')
      elif self.isinDB() and not force:
         criteres = ' AND '.join([
            "%s='%s'" % (k, v) for k, v in self.prim_crit().items() ])
         if __verbose__:
            print "Warning: %s with id=%s is already in database %s (%s)." \
               % (self.__class__.__name__, self.__getid(), self.table, criteres)
      else:
         self.__cachecols()
         new = False
         if not self.isinDB():
            new = True
            if self.get(self.idkey) == None:
               self.__setid(self.__getmaxid()+1)
         dins = self.__encode_string_toDB()
         # écrire les items référencés et les remplacer par leur id
         for k in self.refs.keys():
            if isinstance(self[k], ITEM):
               self[k].write(force_recursive=force_recursive)
               idref = self[k].__getid()
            else:
               if self[k] != None:
                  item = eval('%s(%s)' % (self.refs[k], 'c=self.c'))
                  item.read({item.idkey:self[k]})
               idref = self[k]
            dins[k] = idref
         # dater la modification (activity)
         if dins.has_key('_activity'):
            dins['_activity'] = now
         try:
            if new:
               nb = self.c.insert(self.table, dins)
               self.__setmaxid(self.__getid())
            else:
               nb = self.c.update(self.table, dins, {self.idkey:self.__getid()})
         except MySQLError:
            raise WriteDBError, _('error occurs during writing in database')
         # positionne l'id
         idv = self.__getid_from_db()
         criteres = ' AND '.join([
            "%s='%s'" % (k, v) for k, v in self.prim_crit().items() ])
         if __verbose__:
            print "%s with id=%s written in database %s (%s)." \
               % (self.__class__.__name__, self.__getid(), self.table, criteres)
         # c'est comme si on l'avait lu
         self.lu = True

   def __encode_string_toDB(self):
      """Convert all strings to DB encoding.
      """
      return self.__convert_string(self.std_encoding, self.db_encoding, True)

   def __decode_string_fromDB(self):
      """Convert all strings from DB encoding.
      """
      self.__convert_string(self.db_encoding, self.std_encoding, False)

   def __convert_string(self, in_encoding, out_encoding, copy):
      """Convert all strings from `in_encoding` to `out_encoding`.
      If 'copy' is True a copy of `.values` is returned,
      if 'copy' is False `.values` is modified in place.
      """
      if copy:
         dico = self.values.copy()
      else:
         dico = self.values
      dico = self.values
      for k, v in dico.items():
         if type(v) in StringTypes:
            dico[k] = convert(v, in_encoding, out_encoding)
      if copy:
         return dico

   def __getid_from_db(self, dcrit=None):
      """Retourne l'id de l'enregistrement répondant aux critères fournis
      et positionne le champ id."""
      if self.c == None:
         raise ConnectDBError, _('no database connection')
      else:
         if dcrit == None:
            dcrit = self.prim_crit()
         criteres = ' AND '.join(["%s='%s'" % (k, v) for k, v in dcrit.items()])
         query = """SELECT %s FROM %s WHERE %s;""" \
               % (self.idkey, self.table, criteres)
         res = self.c.exe(query)
         if len(res) == 0:
            raise ReadDBError, _('no item found for %s') % criteres
         elif len(res)>1:
            lid = [l[0] for l in res]
            raise ReadDBError, _('more than one item for %s : %s') \
               % (criteres, repr(lid))
         else:
            self.__setid(res[0][0])
            return self.__getid()

   def __getmaxid(self):
      """Retourne l'id max de la table."""
      if self.c == None:
         raise ConnectDBError, _('no database connection')
      else:
         query = """SELECT max(%s) FROM %s;""" % (self.idkey, self.table)
         res = self.c.exe(query)
         maxi = res[0][0]
         if maxi == None:
            maxi = 0
         try:
            maxi = int(maxi)
         except ValueError:
            raise ReadDBError, _("field '%s' should be numeric") % repr(self.idkey)
         return maxi

   def __setmaxid(self, maxi):
      """Set the max id + 1 in the 'ids' table (for Roundup-style databases).
      """
      if not hasattr(self, 'tab_idmax'):
         return
      if self.c == None:
         raise ConnectDBError, _('no database connection')
      else:
         try:
            self.c.update(self.tab_idmax, {'num' : maxi + 1},
                                          {'name' : re.sub('^[_]+', '', self.table)})
         except MySQLError:
            raise WriteDBError, _('error occurs during writing in database')
   
   def __cachecols(self):
      """Remplit .cols, devrait être appelé par toute méthode l'utilisant."""
      # Si pas déjà rempli
      if self.cols == None:
         if self.c == None:
            raise ConnectDBError, _('no database connection')
         else:
            self.cols = self.c.showcol(self.table)

   def GetLinks(self):
      """Return a list of the items linked to this item (list of versions
      for a product, list of messages for an issue).
      """
      l_items = []
      # ITEM with multi-link
      if self.link_class == None or self.c == None:
         return l_items
      link_table = eval("%s.table" % self.link_class)
      linked_classname = eval("%s.refs['linkid']" % self.link_class)
      linked_class = eval(linked_classname)
      idkey = linked_class.idkey
      query = """SELECT linkid FROM %s WHERE nodeid=%s;""" \
              % (link_table, self.get(self.idkey))
      try:
         res = self.c.exe(query)
      except MySQLError:
         return l_items
      res = [i[0] for i in res]
      if len(res) > 0:
         crit = ' OR '.join(["%s=%s" % (idkey, i) for i in res])
         l_items = ReadDB(linked_class, c=self.c, crit=crit)
      return l_items

   def GetParent(self):
      """Return the parent of this item (a product for a version, an issue for
      a message).
      """
      parent = None
      if self.parent_class == None:
         return parent
      pclass = eval(self.parent_class)
      parent_table = pclass.table
      link_table = eval("%s.table" % pclass.link_class)
      query = """SELECT nodeid FROM %s WHERE linkid=%s;""" \
              % (link_table, self.get(self.idkey))
      try:
         numf = self.c.exe(query)[0][0]
      except (MySQLError, IndexError):
         return parent
      parent = pclass({pclass.idkey : numf}, self.c)
      parent.read()
      return parent
      
   def AsDict(self):
      """Retourne un dictionnaire avec toutes les valeurs, valeurs du champ
      "primaire" pour les sous-items.
      """
      def repr_val(val):
         v = val
         if type(val) in StringTypes:
            mat = re.search(re_date, val)
            if mat != None:
               v = mat.group(1)
         return v
      
      dico = {}
      for k, v in self.values.items():
         if k.startswith('__'):
            continue
         if k in self.refs.keys():
            item = None
            if isinstance(self[k], ITEM):
               item = v
               idref = v.get(v.idkey)
            else:
               item = eval('%s(%s)' % (self.refs[k], 'c=self.c'))
               try:
                  item.read({item.idkey:v})
                  idref = v
               except (ConnectDBError, ReadDBError), msg:
                  item = None
                  if __verbose__:
                     print msg
            if item != None and (item.__class__.__name__ != self.__class__.__name__\
                               or idref != self.get(self.idkey)):
               dico[k] = repr_val(item.GetPrimValue())
         elif not v is None:
            dico[k] = repr_val(v)
      return dico


# ------------------------------------------------------------------------------
def ReadDB(subclass, c, crit=None, verbose=False):
   """Fonction pour lire tous les items de la table associée à subclass.
   """
   if len(subclass.primkeys)<1:
      raise AttributeError, _("no primary keys defined for '%s'") % subclass.__name__
   if crit == None:
      crit = '1'
   query = """SELECT %s FROM %s WHERE %s;""" % \
      (','.join(subclass.primkeys), subclass.table, crit)
   res = c.exe(query)
   ditem = {}
   for lig in res:
      dini = {}
      lkv = transpose([subclass.primkeys, lig])
      for k, v in lkv:
         dini[k] = v
      item = subclass(dini, c=c)
      item.read()
      ditem[lig[0]] = item
   if verbose:
      print '%6d items read from %s' % (len(ditem.keys()), subclass.table)
   return ditem

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
class USER(ITEM):
   """Enregistrement de la table _user"""
   table = '_user'
   default = {
      '_activity'       : now,
      '_actor'          : 1,
      '_address'        : None,
      '_creation'       : now,
      '_creator'        : 1,              # id de admin
      '_loginaster'     : None,
      '_organisation'   : None,
      '_password'       : '{SHA}'+sha.sha('Compaq').hexdigest(),
      '_phone'          : None,
      '_realname'       : None,
      '_roles'          : None,
      '_timezone'       : None,
      '_username'       : None,
      '__retired__'     : 0,
   }
   refs = {}  # pas d'intérêt de mettre _creator car toujours admin, id=1
   primkeys = ['_username', ]

class INTERVENANT(ITEM):
   """Enregistrement de la table _intervenant"""
   table = '_intervenant'
   default = {
      '_activity'       : now,
      '_actor'          : 1,
      '_creation'       : now,
      '_creator'        : 1,              # id de admin
      '_name'           : None,
      '_order'          : None,
      '__retired__'     : 0,
   }
   refs = {}
   primkeys = ['_name', ]

class MSG(ITEM):
   """Enregistrement de la table _msg"""
   table = '_msg'
   parent_class = 'ISSUE'
   default = {
      'id'              : None,
      '_activity'       : now,
      '_actor'          : 1,
      '_author'         : None,
      '_creation'       : now,
      '_creator'        : None,
      '_date'           : now,
      '_inreplyto'      : None,
      '_messageid'      : None,
      '_summary'        : None,  # 255 premiers caractères du message
      '__retired__'     : 0,
   }
   refs = {'_author':'USER', '_creator':'USER', }
   primkeys = ['id', ]

class PRODUIT(ITEM):
   """Enregistrement de la table _produit"""
   table = '_produit'
   link_class = 'PRODUIT_VERSION'
   default = {
      '_activity'       : now,
      '_actor'          : 1,
      '_creation'       : now,
      '_creator'        : 1,
      '_name'           : None,
      '_order'          : None,
      '__retired__'     : 0,
   }
   refs = {}
   primkeys = ['_name', ]

class VERSION(ITEM):
   """Enregistrement de la table _version"""
   table = '_version'
   parent_class = 'PRODUIT'
   default = {
      '_activity'       : now,
      '_actor'          : 1,
      '_creation'       : now,
      '_creator'        : 1,
      '_name'           : None,
      '_order'          : None,
      '__retired__'     : 0,
   }
   refs = {}
   primkeys = ['_name', ]

class PRODUIT_VERSION(ITEM):
   """Enregistrement de la table produit_version"""
   table = 'produit_version'
   default = {
      'linkid'          : None,
      'nodeid'          : None,
   }
   refs = {'linkid':'VERSION', 'nodeid':'PRODUIT'}
   primkeys = ['linkid', ]
   idkey = 'linkid'

class PROJET(ITEM):
   """Enregistrement de la table _projet"""
   table = '_projet'
   default = {
      '_activity'       : now,
      '_actor'          : 1,
      '_creation'       : now,
      '_creator'        : 1,
      '_name'           : None,
      '_order'          : None,
      '__retired__'     : 0,
   }
   refs = {}
   primkeys = ['_name', ]

class STATUS(ITEM):
   """Enregistrement de la table _status"""
   table = '_status'
   default = {
      '_activity'       : now,
      '_actor'          : 1,
      '_creation'       : now,
      '_creator'        : 1,
      '_modifier'       : 1,
      '_name'           : None,
      '_order'          : None,
      '_transition'     : None,
      '__retired__'     : 0,
   }
   refs = {}
   primkeys = ['_name', ]

class TYPE(ITEM):
   """Enregistrement de la table _type"""
   table = '_type'
   default = {
      '_activity'       : now,
      '_actor'          : 1,
      '_creation'       : now,
      '_creator'        : 1,
      '_name'           : None,
      '_order'          : None,
      '__retired__'     : 0,
   }
   refs = {}
   primkeys = ['_name', ]

class ISSUE(ITEM):
   """Enregistrement de la table _issue"""
   table = '_issue'
   link_class = 'ISSUE_MESSAGES'
   default = {
      'id'              : None,
      '_activity'       : now,
      '_actor'          : 1,
      '_assignedto'     : None,
      '_chiffrageCharge': None,
      '_chiffrageDelai' : None,
      '_corrVdev'       : None,
      '_corrVexpl'      : None,
      '_creation'       : now,
      '_creator'        : None,
      '_etatIntervenant': None,
      '_fauxVdev'       : None,
      '_fauxVexpl'      : None,
      '_fichetude'      : None,
      '_impactDoc'      : None,
      '_intervenant'    : None,
      '_nbJours'        : None,
      '_produit'        : None,
      '_projet'         : None,
      '_realiseCharge'  : None,
      '_realiseDelai'   : None,
      '_status'         : None,
      '_title'          : None,
      '_type'           : None,
      '_validReal'      : None,
      '_validation'     : None,
      '_verCorrVdev'    : None,
      '_verCorrVexpl'   : None,
      '_versCible'      : None,
      '_version'        : None,
      '__retired__'     : 0,
   }
   refs = {'_assignedto':'USER', '_creator':'USER',  '_intervenant':'INTERVENANT',
         '_produit':'PRODUIT', '_projet':'PROJET',
         '_status':'STATUS',   '_type':'TYPE',     '_version':'VERSION',
         '_versCible':'VERSION'}
   primkeys = ['id', ]

class ISSUE_MESSAGES(ITEM):
   """Enregistrement de la table issue_messages"""
   table = 'issue_messages'
   default = {
      'linkid'          : None,
      'nodeid'          : None,
   }
   refs = {'linkid':'MSG', 'nodeid':'ISSUE'}
   primkeys = ['linkid', ]
   idkey = 'linkid'





# -*- coding: iso-8859-1 -*-

# ==============================================================================
# COPYRIGHT (C) 1991 - 2003  EDF R&D                  WWW.CODE-ASTER.ORG
# THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
# IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
# THE FREE SOFTWARE FOUNDATION; EITHER VERSION 2 OF THE LICENSE, OR
# (AT YOUR OPTION) ANY LATER VERSION.
#
# THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
# WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. SEE THE GNU
# GENERAL PUBLIC LICENSE FOR MORE DETAILS.
#
# YOU SHOULD HAVE RECEIVED A COPY OF THE GNU GENERAL PUBLIC LICENSE
# ALONG WITH THIS PROGRAM; IF NOT, WRITE TO EDF R&D CODE_ASTER,
#    1 AVENUE DU GENERAL DE GAULLE, 92141 CLAMART CEDEX, FRANCE.
# ==============================================================================

"""
This module gives functions to manipulate Code_Aster jobs :
- return informations from astk server configuration,
- edit output or error file,
- get job status,
- kill and delete the job files,
- search strings in output of a job,
- purge 'flasheur' directory.
The functions are called by an ASTER_RUN object.
"""

__revision__ = "$Id: as_job.py 2627 2006-12-18 07:50:58Z courtois $"


# ----- differ messages translation
def _(mesg):
   return mesg

import os
import signal
import glob
import re
import getpass

from as_profil import ASTER_PROFIL

#-------------------------------------------------------------------------------
def SetParser(run):
   """Configure the command-line parser, add options name to store to the list,
   set actions informations.
   run : ASTER_RUN object which manages the execution
   """
   acts_descr = {
      'info' : {
         'method' : Info,
         'syntax' : '',
         'help'   : 'returns informations from astk server configuration : '\
                    'batch, interactive (yes/no, limits), compute nodes, '\
                    'versions'
      },
      'actu' : {
         'method' : Actu,
         'syntax' : 'job_number job_name mode',
         'help'   : 'returns the state, diagnosis, execution node, spent '\
                    'cpu time and working directory of a job'
      },
      'edit' : {
         'method' : Edit,
         'syntax' : 'job_number job_name mode output|error DISPLAY',
         'help'   : 'opens output or error file on the provided display'
      },
      'del' : {
         'method' : Del,
         'syntax' : 'job_number job_name mode [node] [--signal=...]',
         'help'   : 'kill a job and delete related files'
      },
      'purge_flash' : {
         'method' : Purge,
         'syntax' : 'job_number1 [job_number2 [...]]]',
         'help'   : 'delete files of jobs which are NOT in the list'
      },
      'tail' : {
         'method' : Tail,
         'syntax' : 'job_number job_name mode fdest nb_lines [regexp]',
         'help'   : 'output the last part of fort.6 file or filter lines ' \
                    'matching a pattern'
      },
   }
   opts_descr = {
      'signal' : {
         'args'   : ('--signal', ),
         'kwargs' : {
            'action'  : 'store',
            'default' : 'KILL',
            'choices' : ('KILL', 'USR1'),
            'dest'    : 'signal',
            'help'    : 'signal to the job (KILL|USR1)'
         }
      },
   }
   title = 'Options for operations on jobs'
   run.SetActions(
         actions_descr = acts_descr,
         actions_order = ['info', 'actu', 'edit', 'tail', 'del', 'purge_flash'],
         group_options=True, group_title=title, actions_group_title=False,
         options_descr = opts_descr,
   )

#-------------------------------------------------------------------------------
def Info(run, *args):
   """Return informations from astk server configuration.
   """
   if len(args)>0:
      run.parser.error(_("'--%s' requires no argument") % run.current_action)
   
   on_machref = run.get('rep_agla', 'local') != 'local'
   
   # astk server version
   l_v = run.__version__.split('.')
   try:
      svers = '.'.join(['%02d' % int(i) for i in l_v])
   except ValueError:
      svers = run.__version__
   output = ["@SERV_VERS@", svers, "@FINSERV_VERS@"]

   # all these parameters are necessary   
   output.append("@PARAM@")
   output.extend(['%s : %s' % (p, run[p]) for p in ('plate-forme', 'batch',
      'batch_memmax', 'batch_tpsmax', 'batch_nbpmax', 'interactif',
      'interactif_memmax', 'interactif_tpsmax', 'interactif_nbpmax')])
   output.append("@FINPARAM@")

   # Code_Aster versions
   output.append("@VERSIONS@")
   output.extend(['vers : %s' % v for v in run.get('vers', '').split()])
   output.append("@FINVERSIONS@")

   output.append("@NOEUDS@")
   output.extend(['noeud : %s' % n for n in run.get('noeud', '').split()])
   output.append("@FINNOEUDS@")
   
   if on_machref:
      output.extend(["@MACHREF@", "@FINMACHREF@"])
      output.extend(["@REX_URL@", run.get('rex_url', ''), "@FINREX_URL@"])
      output.extend(["@REX_REPFICH@", run.get('rep_rex', ''), "@FINREX_REPFICH@"])
      output.extend(["@MAIL_ATA@", run.get('mail_ata', ''), "@FINMAIL_ATA@"])
      
      # read identAster file
      f_ident = os.path.join(run['rep_identAster'], 'identAster')
      if not os.path.isfile(f_ident):
         run.Sortie(4)
      f = open(f_ident, 'r')
      l_ident = [line.strip().replace('\t', ' ') \
               for line in f if re.search('^[ ]*%', line)==None]
      f.close()
      d_user = {}
      for l in l_ident:
         l_i = l.split()
         if len(l_i) < 2:
            continue
         d_user[l_i[1]] = l_i
      
      # extract user info
      role = 'UA'
      details = ''
      user = getpass.getuser()
      if d_user.has_key(user):
         l = d_user[user]
         if len(l) >= 6:
            role = l[5].replace('/', ' ')
            details = ('%s %s %s %s' % (l[0], l[1], l[2], l[3])).replace('/', ' ')
      
      output.extend(["@ROLE@", role, "@FINROLE@"])
      if details:
         output.extend(["@INFOID@", details, "@FININFOID@"])

   # message of the day
   output.append("@MOTD@")
   if run.get('motd', '') != '' and os.path.exists(run['motd']):
      output.append(open(run['motd'], 'r').read())
   output.append("@FINMOTD@")
   
   print os.linesep.join(output)

#-------------------------------------------------------------------------------
def _nodepara(run, nodename, para):
   """Return of `para` from config file or config_`nodename` file if exists.
   """
   val = run[para]
   fcn = os.path.join(run.confdir, 'config_%s' % nodename)
   if os.path.isfile(fcn):
      mdict = {}
      run._read_rc(fcn, mdict)
      val = mdict.get(para, val)
   return val

#-------------------------------------------------------------------------------
def _sumtps(l_scpu):
   """Return the sum in seconds of the times given as strings in `l_scpu`.
   Format of each times in [dd:[hh:[mm:[ss]]]]
   """
   tot = 0
   if len(l_scpu) < 1:
      tot = '_'
   for scpu in l_scpu:
      try:
         s = [int(s) for s in scpu.split(':')]
      except ValueError:
         s = []
      if   len(s) == 4:
         tot += 86400*s[0]+3600*s[1]+60*s[2]+s[3]
      elif len(s) == 3:
         tot += 3600*s[0]+60*s[1]+s[2]
      elif len(s) == 2:
         tot += 60*s[0]+s[1]
      elif len(s) == 1:
         tot += s[0]
   return tot

#-------------------------------------------------------------------------------
def Func_actu(run, *args):
   """Return state, diagnosis, node, cpu time and working directory of a job.
   """
   if len(args) != 3:
      run.parser.error(_("'--%s' takes exactly %d arguments (%d given)") % \
         (run.current_action, 3, len(args)))
   
   njob, nomjob, mode = args
   # defaults
   etat  = '_'
   diag  = '_'
   node  = '_'
   tcpu  = '_'
   wrk   = '_'
   queue = '_'
   psout = ''
   on_machref = run.get('rep_agla', 'local') != 'local'
   # the real job id may differ
   jobid = str(njob)
   
   # 1. get information about the job
   # 1.1. batch mode
   if mode == "batch":
      m = 'b'
      # for LSF
      if run['batch_nom'] == 'LSF':
         cmd = ''
         if run.get('batch_ini', '') != '':
            cmd = '. %s ; ' % run['batch_ini']
         cmd += '%s -uall' % run['batch_job']
         jret, out = run.Shell(cmd)
         mat = re.search('(^ *%s.*)' % njob, out, re.MULTILINE)
         if mat == None:
            etat = "ENDED"
         else:
            lin = mat.group(1).split()
            if len(lin) >= 3:
               etat = lin[2]
               if etat != "PEND":
                  node = lin[5]
            if etat.find('SUSP')>-1:
               etat = 'SUSPENDED'
      # for PBS
      elif run['batch_nom'] == 'PBS':
         cmd = ''
         if run.get('batch_ini', '') != '':
            cmd = '. %s ; ' % run['batch_ini']
         cmd += '%s -f %s' % (run['batch_job'], njob)
         jret, out = run.Shell(cmd)
         # real job id
         mjid = re.search('Job Id: *(\S+)', out, re.MULTILINE)
         if mjid != None:
            jobid = mjid.group(1)
         # job state
         metat = re.search(' job_state *= *(\S+)', out, re.MULTILINE)
         if metat == None:
            etat = "ENDED"
         else:
            etat = { 'R' : 'RUN', 'E'  :'RUN',
                     'W' : 'PEND', 'T' : 'PEND', 'Q' : 'PEND',
                     'S' : 'SUSPENDED', 'H' : 'SUSPENDED',
                   }.get(metat.group(1), '?')
         # queue
         mqueue = re.search(' queue *= *(\S+)', out, re.MULTILINE)
         if mqueue != None:
            queue = mqueue.group(1)
         # exec host
         mnode = re.search(' exec_host *= *(\S+)/', out, re.MULTILINE)
         if mnode != None:
            node = mnode.group(1)
      # for Sun Grid Engine
      elif run['batch_nom'] == 'SunGE':
         cmd = ''
         if run.get('batch_ini', '') != '':
            cmd = '. %s ; ' % run['batch_ini']
         cmd += run['batch_job']
         jret, out = run.Shell(cmd)
         mat = re.search('(^ *%s.*)' % njob, out, re.MULTILINE)
         if mat == None:
            etat = "ENDED"
         else:
            lin = mat.group(1).split()
            if len(lin) >= 5:
               etat = lin[4]
               if re.search('[wh]+', etat) != None:
                  etat = "PEND"
               elif re.search('[sST]+', etat) != None:
                  etat = "SUSPENDED"
               else:
                  etat = "RUN"
               if len(lin) >= 8:
                  l_q = lin[7].split('@')
                  queue = l_q[0]
                  if len(l_q) > 1:
                     node  = l_q[1].split('.')[0]
            # because I don't known the running node on which to call 'ps'
            if etat == 'RUN' and node in ('', '_'):
               cmd = ''
               if run.get('batch_ini', '') != '':
                  cmd = '. %s ; ' % run['batch_ini']
               cmd += '%s -j %s' % (run['batch_job'], njob)
               jret, out = run.Shell(cmd)
               expr = re.compile('^usage *[0-9]* *: *cpu=([0-9:]+)', re.MULTILINE)
               l_field = expr.findall(out)
               tcpu = _sumtps(l_field)
      else:
         run.Mess(_('unknown batch scheduler'), _('<F>_ERROR'))

   # 1.2. interactive mode
   elif mode == "interactif":
      m = 'i'
      # astk profile
      pr_astk = os.path.join(run['flasheur'], '%s.p%s' % (nomjob, njob))
      # if it doesn't exist the job is ended
      etat = "ENDED"
      if os.path.isfile(pr_astk):
         prof = ASTER_PROFIL(pr_astk, run)
         node = prof['noeud'][0]
   
   else:
      run.Mess(_('unexpected mode : %s') % mode, _('<F>_UNEXPECTED_VALUE'))

   # 2. query the process
   if node != '_':
      if node.split('.')[0] == run.GetHostName().split('.')[0]:
         jret, psout = run.Shell(run['ps_cpu'])
      else:
         jret, psout = run.Shell(run['ps_cpu'], mach=node)
      # ended ?
      if mode == "interactif" and psout.find('btc.%s' % njob)>-1:
         etat = "RUN"

   # 3.1. the job is ended
   if etat == "ENDED":
      fdiag = os.path.join(run['flasheur'], '%s.%s%s' % (nomjob, m, njob))
      if os.path.isfile(fdiag):
         diag = open(fdiag, 'r').read().split(os.linesep)[0]
         if diag == '':
            diag = '?'
      if diag == '?' :
         diag = _('<F>_SYSTEM')
         # try to find something in output
         fout = os.path.join(run['flasheur'], '%s.o%s' % (nomjob, njob))
         if os.path.isfile(fout):
            f = open(fout, 'r')
            for line in f:
               if line.find('--- DIAGNOSTIC JOB :')>-1:
                  diag = line.split()[4]
               elif line.find('Cputime limit exceeded')>-1:
                  diag = _('<F>_CPU_LIMIT_SYSTEM')
            f.close()
         # working directory
         wrk = _nodepara(run, node, 'rep_trav')
         # copy fort.6 to '.o'
         if not node in ('', '_'):
            ftcp = os.path.join(run['tmp_user'], 'as_actu.%s' % os.getpid())
            fort6 = os.path.join(wrk, '%s.%s.fort.6.%s' % (nomjob, njob, m))
            jret = run.Copy(ftcp, '%s:%s' % (node, fort6), niverr='SILENT')
            if os.path.isfile(ftcp):
               txt = [os.linesep*2]
               txt.append('='*48)
               txt.append('===== Pas de diagnostic, recopie du fort.6 =====')
               txt.append('='*48)
               txt.append(open(ftcp, 'r').read())
               txt.append('='*48)
               txt.append('='*48)
               txt.append(os.linesep*2)
               f = open(fout, 'a')
               f.write(os.linesep.join(txt))
               f.close()
   
   else:
   # 3.2. job is running
      if etat in ('RUN', 'SUSPENDED'):
         # working directory
         wrk = _nodepara(run, node, 'rep_trav')
         if not on_machref or mode == 'interactif':
            wrk = os.path.join(wrk, '%s.%s' % (mode, njob))
         else:
            wrk = os.path.join(wrk, str(jobid))
      if etat == 'RUN' and tcpu == '_':
         # tcpu may have been retrieved upper
         l_tcpu = []
         for line in psout.split(os.linesep):
            if re.search('\-num_job +%s' % njob, line) != None and \
               re.search('\-mode +%s' % mode, line) != None and \
               (not on_machref or line.find(run['rep_agla'])<0):
               l_tcpu.append(re.sub('\..*$', '', line.split()[0]).replace('-', ':'))
         if len(l_tcpu)>0:
            tcpu = _sumtps(l_tcpu)
   
   # 4. return the result
   return etat, diag, node, tcpu, wrk, queue
   
#-------------------------------------------------------------------------------
def Actu(run, *args):
   """Return state, diagnosis, node, cpu time and working directory of a job.
   """
   print "ETAT=%s DIAG=%s EXEC=%s TCPU=%s REP_TRAV=%s QUEUE=%s" % Func_actu(run, *args)

#-------------------------------------------------------------------------------
def Edit(run, *args):
   """Open output or error file of a job.
   """
   if len(args) != 5:
      run.parser.error(_("'--%s' takes exactly %d arguments (%d given)") % \
         (run.current_action, 5, len(args)))

   njob, nomjob, mode, typ, displ = args
   
   # get editor command line
   edit = run['editeur']
   
   # filename to edit
   fich = os.path.join(run['flasheur'], '%s.%s%s' % (nomjob, typ[0], njob))
   
   # open the file
   if os.path.isfile(fich):
      if re.search('display +@D', edit) != None:
         edit = edit.replace('@D', displ)
         kret, out = run.Shell('%s %s' % (edit, fich), bg=True)
      else:
         kret, out = run.Shell('%s -display %s %s' % (edit, displ, fich), bg=True)
   else:
      run.Sortie(4)

#-------------------------------------------------------------------------------
def Del(run, *args):
   """Kill a job and delete related files.
   """
   if len(args) < 3:
      run.parser.error(_("'--%s' takes at least %d arguments (%d given)") % \
         (run.current_action, 3, len(args)))
   elif len(args) > 4:
      run.parser.error(_("'--%s' takes at most %d arguments (%d given)") % \
         (run.current_action, 4, len(args)))
   
   # 0. arguments
   njob, nomjob, mode = args[:3]
   if len(args) > 3:
      node = args[3]
   else:
      node = ''
   
   use_batch_cmd = (mode == "batch" and run['signal'] == 'KILL')
      
   # 1. retrieve the job status
   etat, diag, node, tcpu, wrk, queue = Func_actu(run, njob, nomjob, mode)
   
   # 2. send the signal
   if etat in ('RUN', 'SUSPENDED', 'PEND'):
      if use_batch_cmd:
         if run['batch_nom'] in ('LSF', 'PBS', 'SunGE'):
            cmd = ''
            if run.get('batch_ini', '') != '':
               cmd = '. %s ; ' % run['batch_ini']
            cmd += '%s %s' % (run['batch_kil'], njob)
            iret, out = run.Shell(cmd)
            if iret != 0:
               run.Sortie(4)
         else:
            run.Mess(_('unknown batch scheduler'), _('<F>_ERROR'))
      else:
         numpr = ''
         # get process id
         if node != '_':
            if node.split('.')[0] == run.GetHostName().split('.')[0]:
               jret, psout = run.Shell(run['ps_pid'])
            else:
               jret, psout = run.Shell(run['ps_pid'], mach=node)
            exp = re.compile('^ *([0-9]+) +(.*)\-num_job +%s.*\-mode +%s' % (njob, mode),
                  re.MULTILINE)
            res = exp.findall(psout)
            for numj, cmd in res:
               if cmd.find('/bin/sh') < 0:
                  numpr = int(numj)
                  break
         if numpr != '':
            if node.split('.')[0] == run.GetHostName().split('.')[0]:
               os.kill(numpr, getattr(signal, 'SIG%s' % run['signal']))
            else:
               iret, psout = run.Shell('kill -%s %s' % (run['signal'], numpr), mach=node)
         else:
            run.Mess(_('<as_job.Del> process found :'))
            run.Mess(psout, 'SILENT')
            run.Mess(_('process not found'), _('<F>_ERROR'))
   
   # 3. delete files
   if run['signal'] == 'KILL':
      l_fich = glob.glob(os.path.join(run['flasheur'], '%s.?%s' % (nomjob, njob)))
      for f in l_fich:
         run.Delete(f)
   

#-------------------------------------------------------------------------------
def Tail(run, *args):
   """Output the last part of fort.6 file or filter lines matching a pattern.
   """
   if len(args) < 5:
      run.parser.error(_("'--%s' takes at least %d arguments (%d given)") % \
         (run.current_action, 5, len(args)))
   elif len(args) > 6:
      run.parser.error(_("'--%s' takes at most %d arguments (%d given)") % \
         (run.current_action, 6, len(args)))
   
   # 0. arguments
   njob, nomjob, mode, fdest, nbline = args[:5]
   if len(args) > 5:
      cmd = 'egrep -- \'%s\' ' % args[5]
   else:
      cmd = 'tail -%s ' % nbline
   
   # 1. retrieve the job status
   etat, diag, node, tcpu, wrk, queue = Func_actu(run, njob, nomjob, mode)

   # 2. fill output file
   s_out = ''
   ftmp = os.path.join(run['tmp_user'], 'as_tail.%s' % os.getpid())
   if mode == 'batch' and run['batch_nom'] == 'SunGE':
      s_out = _("Sorry I don't know how to ask Sun Grid Engine batch " \
                "scheduler the running node.")
   if etat == 'RUN':
      # 2.1.1. file to parse
      fich = os.path.join(wrk, 'fort.6')
      if node != '_':
         mach = node
         fich = '%s:%s' % (node, fich)
      else:
         mach = ''
      # 2.1.2. execute command
      if run.Exists(fich):
         jret, s_out = run.Shell(cmd + run.PathOnly(fich), mach=node)
      else:
         run.Mess(_('file is empty'))
   else:
      # 2.2. job isn't running
      print "JOB=%s JOBID=%s ETAT=%s" % (nomjob, njob, etat)
      run.Sortie(4)
   
   # 3. send output file
   if run.IsRemote(fdest):
      open(ftmp, 'w').write(s_out)
      jret = run.Copy(fdest, ftmp)
   else:
      fdest = run.PathOnly(fdest)
      open(fdest, 'w').write(s_out)

#-------------------------------------------------------------------------------
def Purge(run, *args):
   """Delete files which are NOT listed in args.
   """
   if len(args) < 1:
      run.parser.error(_("'--%s' takes at least %d arguments (%d given)") % \
         (run.current_action, 1, len(args)))
   
   # 0. arguments
   njob = len(args)
   l_job = []
   for j in args:
      try:
         l_job.append(int(j))
      except ValueError:
         run.parser.error(_('incorrect job number : %s') % j)

   # 1. action
   l_keep = []
   l_f = os.listdir(run['flasheur'])
   for f in l_f:
      delete = True
      for j in l_job:
         if re.search('\.[a-z]?%d$' % j, f) != None:
            delete = False
            break
      if delete:
         run.Delete(os.path.join(run['flasheur'], f), verbose=True)




# -*- coding: iso-8859-1 -*-

# ==============================================================================
# COPYRIGHT (C) 1991 - 2003  EDF R&D                  WWW.CODE-ASTER.ORG
# THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
# IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
# THE FREE SOFTWARE FOUNDATION; EITHER VERSION 2 OF THE LICENSE, OR
# (AT YOUR OPTION) ANY LATER VERSION.
#
# THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
# WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. SEE THE GNU
# GENERAL PUBLIC LICENSE FOR MORE DETAILS.
#
# YOU SHOULD HAVE RECEIVED A COPY OF THE GNU GENERAL PUBLIC LICENSE
# ALONG WITH THIS PROGRAM; IF NOT, WRITE TO EDF R&D CODE_ASTER,
#    1 AVENUE DU GENERAL DE GAULLE, 92141 CLAMART CEDEX, FRANCE.
# ==============================================================================

"""Definition of ASTER_RUN class.
"""

__revision__ = "$Id: as_run.py 2686 2006-12-22 16:55:38Z courtois $"
__svntag__ = "$HeadURL: svn://aster.cla.edfgdf.fr/astk/tags/v1.5.0/ASTK_SERV/lib/as_run.py $"

# ----- differ messages translation
def _(mesg):
   return mesg

import sys
import os
import re
import time
import getpass
import traceback
import pprint
from types import StringType, DictType, ListType, TupleType, StringTypes
from optparse import OptionParser, OptionGroup, SUPPRESS_HELP

# ----- import as_run modules
from as_option    import ASTER_OPTION
from utils        import get_absolute_dirname

# ----- version décodée du tag cvs
__version__ = 'VDEV'
# ----- valeurs CVS
_mat = re.search('astk/tags/v([\._0-9rc]+)/ASTK_SERV', __svntag__)
if _mat != None:
   __version__ = _mat.group(1).replace('_', '.')

__copyright__ = 'Copyright (c) 2001-'+time.strftime('%Y')+ \
      ' EDF R&D - http://www.code-aster.org'

class RunAsterError(Exception):
   """Base exception for ASTER_RUN
   """

class ASTER_RUN:
   # Main class  (__doc__ is used as help string for the parser)
   """%(program)s action [options] [arguments]

  Functions :
"""
#-------------------------------------------------------------------------------
   def __init__(self, astk_serv_root):
      """Initialisations. 'astk_serv_root' is the main astk directory containing
      'bin', 'lib' and 'conf' directories ([...]/ASTK/ASTK_SERV).
      """
      self.ExitOnFatalError = True
      self.astk_serv_root = astk_serv_root
      self.aster_root     = get_absolute_dirname(get_absolute_dirname(astk_serv_root))
      self.confdir = os.path.join(astk_serv_root, 'conf')
      self.__initialized = False
      self.__dbg_count = 0
      # ----- verification de la plate-forme
      if os.name != 'posix':
         print _("""This program runs only under Linux/Unix system'
... but its license allows you to port it under another system and propose your
modifications to the author !
Porting should be limited to 'os.system' calls and other
functions of the 'os' module which are not platform independent.
Contact : http://www.code-aster.org""")
         self.Sortie(4)

      # ----- verification de la version de Python
      if sys.hexversion < 0x020300F0:
         print _("This script requires Python 2.3 or higher, sorry !")
         self.Sortie(4)

      # ----- nom du programme
      self.program = 'as_run'
      self.__version__ = __version__
      
      # ----- options, liste des options du parser à stocker dans options
      self.options = { 'verbose': False }
      # always use _add_option to extend options_list
      self.options_list = ['debug', 'verbose', 'force', 'num_job']
      
      # ----- informations about actions : module, method
      # always use set_action to define a new action
      self.actions_info = {}
      # store current action
      self.current_action = ''

      # ----- user preferences and defaults
      self.user_rc = os.path.join(os.getenv('HOME'), '.astkrc', 'config')
      self.user_vars = {
         'editor' :
            {  'val'  : 'nedit',
               'help' : 'editor command' },
         'devel_server_user' :
            {  'val'  : '',
               'help' : 'login on the development server' + os.linesep + \
               '# (name/ip address is usually set in ASTK_SERV/conf/config)' },
         'remote_copy_protocol' :
            {  'val'  : 'RCP',
               'help' : 'remote protocol used to copy files and directories' },
         'remote_shell_protocol' :
            {  'val'  : 'RSH',
               'help' : 'remote protocol used for shell commands' },
      }
      self._add_option(*self.user_vars.keys())

      # ----- formats
      self.fmt = {
         'usage'     : '   - %s :'+os.linesep+'      %s'+os.linesep,
         'verb1'     : ' %-15s = %s',
         'msg+cod'   : os.linesep+'%-18s %s'+os.linesep,
         'msg_info'  : '<INFO>%s %s',
         'silent'    : '      %s %s',
         'title'     : os.linesep+'-'*80+os.linesep+'%s %s'+os.linesep,
         'diag'      : os.linesep+'-'*60+os.linesep+\
                       '--- DIAGNOSTIC JOB : %s'+\
                       os.linesep+'-'*60+os.linesep,
         'exit'      : os.linesep+_('<ERROR> EXIT_CODE=')+'%d'
      }

      # ----- diagnostic
      self.diag = ''

      # ----- list of files/directories to delete on exit
      self.to_delete = []

      # ----- lists of text to print on exit depending on diagnostic
      # this allows to keep important messages when a lot of output is generated
      # and to print them once again at end
      self.print_on_exit = {}

      # ----- do not print timer info on exit by default
      self.timer = None
      self.print_timer = False

      # ----- parser
      p = OptionParser(option_class=ASTER_OPTION,
         usage=self.__doc__ % vars(self),
         version='%prog '+self.__version__+os.linesep+__copyright__)
      self.parser = p
      p.add_option('-v', '--verbose',
         action='store_true', dest='verbose', default=self.options['verbose'],
         help=_('print status messages'))
      p.add_option('-g', '--debug',
         action='store_true', dest='debug', default=False,
         help=_('print debugging information'))
      p.add_option('-f', '--force',
         action='store_true', dest='force', default=False,
         help=_('force operations which can be cached (download, ' \
                'compilation...)'))
      p.add_option('--num_job',
         action='store', dest='num_job', default=str(os.getpid()),
         help=SUPPRESS_HELP)

      # ----- add user's preferences as options
      for k, dV in self.user_vars.items():
         p.add_option('--'+k,
            action='store', dest=k, default=None,
            help=dV['help'].replace(os.linesep, ' ').replace('#', ''))

      # ----- read master configuration file and user preferences
      self.config = {}
      self._read_rc(os.path.join(self.confdir, 'config'), self.config)
      self._init_rc()

      # ----- prepare tmp_user and cache_dir
      # no folder in cache_dir
      tmp_user = os.path.join(self.config['rep_tmp'],
            'astk_'+getpass.getuser())
      self.config['tmp_user'] = tmp_user
      self.config['cache_dir'] = os.path.join(tmp_user, 'cache')
      if not os.path.exists(tmp_user):
         os.makedirs(tmp_user)
      elif not os.access(tmp_user, os.W_OK):
         print _('no write access to %s') % tmp_user
         self.Sortie(4)
      
      if not os.path.exists(self.config['cache_dir']):
         os.mkdir(self.config['cache_dir'])

      # ----- check for 'flasheur'
      flash = os.path.join(os.getenv('HOME'), self.config['flash_srv'])
      self.config['flasheur'] = flash
      if not os.path.exists(flash):
         os.makedirs(flash)
      elif not os.access(flash, os.W_OK):
         print _('no write access to %s') % flash
         self.Sortie(4)

#-------------------------------------------------------------------------------
   def ParseArgs(self):
      """Calls the arguments parser, returns options and arguments.
      """
      opts, args = self.parser.parse_args()
      # ----- all should have been initialized !
      self.__initialized = True
      return opts, args

#-------------------------------------------------------------------------------
   def _init_rc(self):
      """Read user preferences or create the file the first time
      """
      ficrc = self.user_rc
      # ----- check for ressources directory
      if not os.path.isdir(os.path.dirname(ficrc)):
         os.makedirs(os.path.dirname(ficrc))
      
      if not os.path.isfile(ficrc):
         # ----- create ressources file
         fic = open(ficrc, 'w')
         fic.write('# User preferences file' + os.linesep*2)
         fic.write('# You can override here all values of ' \
               'ASTK_SERV/conf/config' + os.linesep*2)
         for var in self.user_vars.keys():
            fic.write(('# %s'+os.linesep+'%s : %s' +os.linesep*2) % \
                  (self.user_vars[var]['help'], var, self.user_vars[var]['val']))
         fic.close()
      # ----- read ressource file
      self._read_rc(ficrc, self.config)

#-------------------------------------------------------------------------------
   def _read_rc(self, ficrc, destdict):
      """Read a ressource file and store variables to 'destdict'.
      """
      if os.path.isfile(ficrc):
         f = open(ficrc, 'r')
         for line in f:
            if not re.search('^[ ]*#', line):
               mat = re.search('([-a-z_A-Z0-9]+) *: *(.*)', line.strip())
               if mat:
                  key   = mat.group(1)
                  value = mat.group(2).strip()
                  # try to convert numbers and boolean values
                  try:
                     if value.isdigit():
                        value = int(value)
                     else:
                        value = float(value)
                  except ValueError:
                     if value in ('True', 'False'):
                        value = eval(value)
                  # repetable ?
                  if destdict.has_key(key) and key in ('vers', 'noeud'):
                     value = str(destdict[key])+' '+str(value)
                  destdict[key] = value
         f.close()
      else:
         print _('file not found : %s') % ficrc
         self.Sortie(4)

#-------------------------------------------------------------------------------
   def _add_action(self, action, info):
      """Set informations about an action
      Prefer use this method over extend manually the dictionnary.
      """
      if not type(action) in StringTypes:
         self.Mess(_('invalid type (expect %s not %s)') % \
               ('string', type(action)), _('<F>_PROGRAM_ERROR'))
      if not type(info) is DictType:
         self.Mess(_('invalid type (expect %s not %s)') % \
               ('dict', type(info)), _('<F>_PROGRAM_ERROR'))
      if action in self.actions_info.keys():
         self.Mess(_('action already defined : %s') % action,
               _('<F>_PROGRAM_ERROR'))
      else:
         self.actions_info[action] = info

#-------------------------------------------------------------------------------
   def _add_option(self, *args):
      """Add one or more option which will be available through getitem.
      Prefer use this method over extend manually the list.
      """
      if not type(args) in (ListType, TupleType):
         self.Mess(_('invalid type (expect %s not %s)') % \
               ('string', type(args)), _('<F>_PROGRAM_ERROR'))
      for opt in args:
         if opt in self.options_list:
            self.Mess(_('option already defined : %s') % opt,
               _('<F>_PROGRAM_ERROR'))
         else:
            self.options_list.append(opt)

#-------------------------------------------------------------------------------
   def SetActions(self, actions_descr, **dico):
      """Add new actions to the parser configuration
         actions_descr  : description of each action :
            key   = action name,
            value = {syntax, help, method}
      Optionnal arguments :
         actions_order  : order to display actions in help (default to '' :
            arbitrary order of dico.keys method)
         group_options  : add a option group to list options (default to True)
         group_title    : title of the option group (default to '')
         actions_group_title : list actions names in the title between
            parenthesis (default to True)
         options_descr  : description of each option (default to {}) :
            key   = option name,
            value = {args, kwargs} passed to add_option parser method
         stored_options : list of options which should be available using
            getitem (default to all "dest" from options_descr)
      """
      # ----- default values
      # actions_order
      if dico.has_key('actions_order'):
         lacts = dico['actions_order']
      else:
         lacts = actions_descr.keys()
         lacts.sort()
      # title
      if dico.has_key('group_title'):
         title = dico['group_title']+' ('
      else:
         title = ''
      # actions_group_title
      if not dico.has_key('actions_group_title'):
         dico['actions_group_title'] = True
      #  group_options
      if not dico.has_key('group_options'):
         dico['group_options'] = True
      # options_descr
      if not dico.has_key('options_descr'):
         dico['options_descr'] = {}
      # stored_options
      if not dico.has_key('stored_options'):
         dico['stored_options'] = [d.values()[1]['dest'] \
               for d in dico['options_descr'].values()]
      # ----- check types
      # dict
      for obj in (actions_descr, dico['options_descr']):
         if not type(obj) is DictType:
            self.Mess(_('invalid type (expect %s not %s)') % \
                  ('dict', type(obj)), _('<F>_PROGRAM_ERROR'))
      # list
      for obj in (lacts, dico['stored_options']):
         if not type(obj) is ListType:
            self.Mess(_('invalid type (expect %s not %s)') % \
                  ('list', type(obj)), _('<F>_PROGRAM_ERROR'))
      prem = True
      for act in lacts:
         descr = actions_descr[act]
         # ----- actions_info
         self._add_action(act, { 'method' : descr['method'] })
         # ----- parser
         if descr['help'] != SUPPRESS_HELP:
            new_usage = self.parser.get_usage()+ \
               (self.fmt['usage'] % \
                  (descr['help'].capitalize(), '%prog --'+act+' '+descr['syntax']))
            self.parser.set_usage(new_usage)
         self.parser.add_option('--'+act,
               action='store_const_once', dest='action', const=act,
               help=SUPPRESS_HELP)
         if not prem:
            title = title+', '
         prem = False
         title = title+('%r' % act)
      title = title+') '
      if not dico['actions_group_title']:
         title = dico['group_title']
      # ----- specific options
      if dico['group_options']:
         group = OptionGroup(self.parser, title=title)
         self.parser.add_option_group(group)
      else:
         group = self.parser
      for opt, descr in dico['options_descr'].items():
         group.add_option(*descr['args'], **descr['kwargs'])
      # ----- add options to the options list to store
      self._add_option(*dico['stored_options'])

#-------------------------------------------------------------------------------
   def PrintConfig(self):
      """Imprime la liste et la valeur des paramètres
      """
      self.parser.print_version()
      print os.linesep+_('Parameters (config attribute)')+' :'
      for key, val in self.config.items():
         print self.fmt['verb1'] % (key, val)
      print os.linesep+_('Options (options attribute)')+' :'
      for key, val in self.options.items():
         print self.fmt['verb1'] % (key, val)

#-------------------------------------------------------------------------------
   def _clear_tmp(self):
      """Empty cache directory if necessary (remove files for which last access
      is older than 'deltat' seconds), and delete all directories
      from 'self.to_delete'.
      """
      ct = time.time()
      deltat = 24*3600
      # ----- config is not set during __init__
      if self.__initialized:
         if self['verbose']:
            print _('Clear cache directory')+' '+self['cache_dir']+'...'
         # ----- avoid deleting such folders : /, /usr or /home !!!
         if re.search('^/.+/.*', self['cache_dir']):
            for root, dirs, files in os.walk(self['cache_dir'], topdown=False):
               for name in files:
                  if ct-os.stat(os.path.join(root, name)).st_atime > deltat:
                     os.remove(os.path.join(root, name))
                  for name in dirs:
                     try:
                        os.rmdir(os.path.join(root, name))
                     except OSError:
                        pass
         # ----- delete each directory in list_dirs
         if self['verbose']:
            print _('Clear temporary files/directories')
         # if a Delete system function has been added
         if hasattr(self, 'Delete'):
            for d in self.to_delete:
               getattr(self, 'Delete')(d)

#-------------------------------------------------------------------------------
   def __getitem__(self, key):
      """Méthode pour accéder facilement aux paramètres de configuration ou à la
      valeur d'une option.
      Les options surchargent les paramètres de configuration.
      """
      if globals().has_key('__'+key+'__'):
         return globals()['__'+key+'__']
      elif self.options.get(key) != None:
         return self.options[key]
      elif self.config.get(key) != None:
         return self.config[key]
      elif self.__initialized:
         self.Mess(_("""'%s' not exist in config or options.
 Perhaps it must be defined in %s.""") % (key, self.user_rc), \
               _('<F>_PROGRAM_ERROR'))

#-------------------------------------------------------------------------------
   def get(self, key, default=None):
      """Semblable à getitem avec la possibilité de retourner une valeur par défaut
      si la valeur n'est pas trouvée.
      """
      if globals().has_key('__'+key+'__'):
         return globals()['__'+key+'__']
      elif self.options.get(key) != None:
         return self.options[key]
      elif self.config.get(key) != None:
         return self.config[key]
      else:
         return default

#-------------------------------------------------------------------------------
   def PostConf(self):
      """Ajuste la configuration dynamiquement (car dépendance au contexte et
      pas seulement au fichier 'config').
      Paramètres déclenchants :
         serv_as_node, only_serv_as_node
      """
      if not self.__initialized or not hasattr(self, 'GetHostName'):
         print _('Warning : ASTER_RUN object not ready for dynamic adjustement !')
         return
      
      # add localhost as a compute node
      if self.config.get('serv_as_node', False):
         l_node = self.config.get('noeud', '').split()
         localhost = self.GetHostName()
         localhost_short = localhost.split('.')[0]
         if not localhost in l_node and not localhost_short in l_node:
            l_node.insert(0, localhost_short)
         if self.config.get('only_serv_as_node', False):
            l_node = [localhost_short,]
         self.config['noeud'] = ' '.join(l_node)

#-------------------------------------------------------------------------------
   def GetGrav(self, cod):
      """Return the severity behind 'cod' as a number to allow comparison
      """
      dgrav = { '?' : -9, 'NO_TEST_RESU' : -1, 'OK' : 0, 'A' : 1, 'NOOK' : 2,
              'S' : 4, 'E' : 6, 'NO_RESU_FILE' : 6, 'F' : 10 }
      g = dgrav['?']
      mat = re.search('<(.)>', cod)
      if cod in dgrav.keys():
         g = dgrav[cod]
      elif re.search('NOOK', cod):
         g = dgrav['NOOK']
      elif mat != None:
         try:
            g = dgrav[mat.group(1)]
         except KeyError:
            pass
      return g

#-------------------------------------------------------------------------------
   def Mess(self, msg, cod='', store=False):
      """Print a message sur stdout,
      'cod' is an error code (format "<.>_...")
         <E> : continue,
         <F> : stop,
         <A> : alarm, continue,
         '' or INFO : for an info,
         SILENT : like an info without <INFO> string,
         TITLE : add a separator.
      If cod='<F>' (without a description), if exists last <E> error is
      transformed into <F>, else <F>_ABNORMAL_ABORT.
      If store is True, 'msg' is stored in print_on_exit dictionnary.
      """
      # ----- gravite
      g0 = self.GetGrav(self.diag)
      g1 = self.GetGrav(cod)
      coderr = cod
      if cod == '<F>':
         if g0 < self.GetGrav('<E>'):
            coderr =  _('<F>_ABNORMAL_ABORT')
         else:
            coderr = self.diag.replace('<E>', '<F>')
      if g1 >= self.GetGrav('<A>'):
         self.DBG('Alarme ou erreur déclenchée :', '%s %s' % (cod, msg),
                  print_traceback=True)

      # ----- message
      if cod == '' or cod == 'INFO':
         fmt = self.fmt['msg_info']
         coderr = ''
      elif cod == 'SILENT':
         fmt = self.fmt['silent']
         coderr = ''
      elif cod == 'TITLE':
         fmt = self.fmt['title']
         coderr = ''
      else:
         fmt = self.fmt['msg+cod']
         # unknown flag
         if g1 == -9:
            coderr = '<I> '+coderr
      print fmt % (coderr, msg)
      sys.stdout.flush()
      # ----- store in print_on_exit
      if store:
         k = '?'
         msg2 = msg
         mat = re.search('<(.)>', cod)
         if mat != None and mat.group(1) in ('A', 'S', 'E', 'F'):
            k = mat.group(1)
            msg2 = self.fmt['msg+cod'] % (coderr, msg)
         elif cod in ('OK', 'NOOK'):
            k = cod
         if not self.print_on_exit.has_key(k):
            self.print_on_exit[k] = [msg2]
         else:
            self.print_on_exit[k].append(msg2)

      # ----- diagnostic le plus défavorable
      if g1 > g0:
         self.diag = coderr
         if g1 == self.GetGrav('<F>'):
            self.Sortie(4)

#-------------------------------------------------------------------------------
   def Sortie(self, exit_code):
      """Exit by printing diagnostic if defined or exit code if not null
      """
      # print important messages
      titles = {
         '?'      : _('Important messages previously printed :')+os.linesep,
         'OK'     : _('Successful messages previously printed:')+os.linesep,
         'NOOK'   : _('NOOK messages previously printed:')+os.linesep,
         'A'      : _('<A> Alarms previously raised :')+os.linesep,
         'S'      : _('<S> errors previously raised :')+os.linesep,
         'E'      : _('<E> errors previously raised :')+os.linesep,
         'F'      : _('<F> errors previously raised :')+os.linesep,
      }
      for k, tit in titles.items():
         lmsg = self.print_on_exit.get(k, [])
         if len(lmsg)>0:
            print self.fmt['title'] % (tit, os.linesep.join(lmsg))
      
      # stop all timers
      if self.print_timer and hasattr(self, 'timer'):
         print os.linesep, self.timer

      if self.diag:
         # change <E> to <F>
         self.diag = self.diag.replace('<E>', '<F>')
         print self.fmt['diag'] % self.diag
      elif exit_code != 0:
         print self.fmt['exit'] % exit_code
      self._clear_tmp()
      # helps to locate error in PROGRAM_ERROR case or in verbose mode
      if self.diag == _('<F>_PROGRAM_ERROR') \
            or not self.ExitOnFatalError \
            or (exit_code != 0 and (self['debug'] or self['verbose'])):
         raise RunAsterError, exit_code
      else:
         sys.exit(exit_code)

#-------------------------------------------------------------------------------
   def CheckOK(self, tole='<S>'):
      """Exit one error more severe than 'tole' occurred.
      """
      if self.GetGrav(self.diag) >= self.GetGrav(tole):
         self.Sortie(4)

#-------------------------------------------------------------------------------
   def ReinitDiag(self, diag='?'):
      """Reinitialize diagnostic (for example if a global diagnostic is stored)
      """
      self.diag = diag

#-------------------------------------------------------------------------------
   def _printDBG(self, *args, **kargs):
      """Print debug information to stderr.
      """
      print_traceback = kargs.get('print_traceback', True)
      file            = kargs.get('file', sys.stderr)
      try:
         self.__dbg_count += 1
         form = """
-------------------------------------------------------------------------------
>>> %(num)09d  %(orig)s

%(content)s
%(traceback)s"""
         formTB = """
Traceback:
%s
"""
         stack = traceback.format_stack(limit=10)
         try:
            ls = []
            for a in args:
               if type(a) in StringTypes:
                  ls.append(a)
               else:
                  ls.append(pprint.pformat(a))
            content = os.linesep.join(ls)
            if len(content) > 800:
               content = content[:800] + os.linesep + '[...]'
         except None:
            content = str(args)
         dinfo = {
            'num'       : self.__dbg_count,
            'orig'      : stack[-3],         # -1 : _printDBG, -2 : DBG ou ASSERT
            'content'   : content,
            'traceback' : '',
         }
         try:
            mat = re.search('File [\'\"]*(.*?)[\'\"]*, *line ([0-9]+), *in (.*)', dinfo['orig'])
            dinfo['orig'] = '[%s@%s:%s]' % (mat.group(3), mat.group(1), mat.group(2))
         except None:
            pass
         if print_traceback:
            dinfo['traceback'] = formTB % (''.join(traceback.format_stack(limit=10)))
         file.write(form % dinfo)
         file.flush()
      except None:
         pass

#-------------------------------------------------------------------------------
   def DBG(self, *args, **kargs):
      """Print debug information to stderr.
      """
      print_traceback = kargs.get('print_traceback', False)
      file            = kargs.get('file', sys.stderr)
      if (not file.isatty()) or self['debug'] or self['verbose']:
         self._printDBG(print_traceback=print_traceback, file=file, *args)

#-------------------------------------------------------------------------------
   def ASSERT(self, condition):
      """Print errors to stdout and stderr.
      """
      if not condition:
         for f in (sys.stdout, sys.stderr):
            self._printDBG('Assertion failed', print_traceback=True, file=f)



# -*- coding: iso-8859-1 -*-

# ==============================================================================
# COPYRIGHT (C) 1991 - 2003  EDF R&D                  WWW.CODE-ASTER.ORG
# THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
# IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
# THE FREE SOFTWARE FOUNDATION; EITHER VERSION 2 OF THE LICENSE, OR
# (AT YOUR OPTION) ANY LATER VERSION.
#
# THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
# WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. SEE THE GNU
# GENERAL PUBLIC LICENSE FOR MORE DETAILS.
#
# YOU SHOULD HAVE RECEIVED A COPY OF THE GNU GENERAL PUBLIC LICENSE
# ALONG WITH THIS PROGRAM; IF NOT, WRITE TO EDF R&D CODE_ASTER,
#    1 AVENUE DU GENERAL DE GAULLE, 92141 CLAMART CEDEX, FRANCE.
# ==============================================================================

"""ASTER_BUILD class.
"""

__revision__ = "$Id: as_build.py 2566 2006-12-08 10:09:20Z courtois $"


# ----- differ messages translation
def _(mesg):
   return mesg

import os
import sys
import glob
import re
import copy
import time
import py_compile
from types import ListType, TupleType

fmt_catapy = 'cata%s.py%s'
fcapy  = fmt_catapy % ('', '')
fcapy_ = fmt_catapy % ('', '*')

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
class ASTER_BUILD:
   """This class provides functions to build a version of Code_Aster.
   """
   repcatapy = ('entete', 'commun', 'commande')
   repcatalo = ('compelem', 'typelem', 'options')

#-------------------------------------------------------------------------------
   def __init__(self, run, conf):
      """run   : ASTER_RUN object
      conf : ASTER_CONFIG object
      """
      # initialisations
      self.verbose = False
      self.debug = False
      # ----- reference to ASTER_RUN object which manages the execution
      self.run = run
      if run != None and run.__class__.__name__ == 'ASTER_RUN':
         self.verbose = run['verbose']
         self.debug = run['debug']
      # ----- Check if 'conf' is a valid ASTER_CONFIG object
      self.conf = conf
      if conf == None or conf.__class__.__name__ != 'ASTER_CONFIG':
         self._mess(_('no configuration object provided !'), _('<F>_ERROR'))
      else:
         # ----- check that ASTER_CONFIG fields are correct (not nul !)
         pass

#-------------------------------------------------------------------------------
   def _mess(self, msg, cod='', store=False):
      """Just print a message
      """
      if hasattr(self.run, 'Mess'):
         self.run.Mess(msg, cod, store)
      else:
         print '%-18s %s' % (cod, msg)

#-------------------------------------------------------------------------------
   def Compil(self, typ, rep, repobj, dbg, rep_trav='', error_if_empty=False):
      """Compile 'rep/*.suffix' files and put '.o' files in 'repobj'
         dbg : debug or nodebug
         (rep can also be a file)
      rep can be remote.
      """
      self.run.DBG('type : %s' % typ, 'rep : %s' % rep, 'repobj : %s' % repobj)
      prev = os.getcwd()
      # ----- get command line from conf object
      if typ == 'C':
         suffix = '.c'
         if self.conf['CC'][0] == '':
            self._mess(_("C compiler not defined in 'config.txt' (CC)"), \
                  _('<F>_CMD_NOT_FOUND'))
         cmd = self.conf['CC'][:]
         if dbg == 'debug':
            cmd.extend(self.conf['OPTC_D'])
         else:
            cmd.extend(self.conf['OPTC_O'])
         cmd.append('-I.')
         cmd.extend(self.conf['INCL'])
      elif typ == 'F':
         suffix = '.f'
         if self.conf['F77'][0] == '':
            self._mess(_("Fortran 77 compiler not defined in 'config.txt' (F77)"), \
                  _('<F>_CMD_NOT_FOUND'))
         cmd = self.conf['F77'][:]
         if dbg == 'debug':
            cmd.extend(self.conf['OPTF_D'])
         else:
            cmd.extend(self.conf['OPTF_O'])
         cmd.append('-I.')
         cmd.extend(self.conf['INCLF'])
      elif typ == 'F90':
         suffix = '.f'
         if self.conf['F90'][0] == '':
            self._mess(_("Fortran 90 compiler not defined in 'config.txt' (F90)"), \
                  _('<F>_CMD_NOT_FOUND'))
         cmd = self.conf['F90'][:]
         if dbg == 'debug':
            cmd.extend(self.conf['OPTF90_D'])
         else:
            cmd.extend(self.conf['OPTF90_O'])
         cmd.append('-I.')
         cmd.extend(self.conf['INCLF90'])
      else:
         self._mess(_('unknown type : %s') % typ, _('<F>_PROGRAM_ERROR'))
      cmd = ' '.join(cmd)
      
      # ----- if force is True, don't use ctime to skip a file
      force = self.run['force']
      
      # ----- copy source files if remote
      if self.run.IsRemote(rep):
         if rep_trav == '':
            self._mess(_('rep_trav must be defined'), _('<F>_PROGRAM_ERROR'))
         obj = os.path.join(rep_trav, '__tmp__'+os.path.basename(rep))
         iret = self.run.MkDir(obj)
         if self.run.IsDir(rep):
            src = [os.path.join(rep, '*'+suffix), os.path.join(rep, '*.h')]
         else:
            src = [rep,]
         iret = self.run.Copy(obj, niverr='SILENT', *src)
         rep = obj
      else:
         rep = self.run.PathOnly(rep)

      # ----- check that directories exist
      if not os.path.isdir(repobj):
         iret = self.run.MkDir(repobj, verbose=True)
      
      # ----- work in repobj
      os.chdir(repobj)

      # ----- log file
      msg = os.path.basename(rep)+'.msg'
      if os.path.exists(msg):
         os.remove(msg)
      fmsg = open(msg, 'w')
      fmsg.write(os.linesep + \
                 'Messages de compilation' + os.linesep + \
                 '=======================' + os.linesep)

      # ----- list of source files
      if self.run.IsDir(rep):
         files = glob.glob(os.path.join(rep, '*'+suffix))
      else:
         files = [rep, ]
      if len(files) == 0:
         iret = 0
         niverr = ''
         if error_if_empty:
            iret = 2
            niverr = _('<E>_NO_SOURCE_FILE')
         self._mess(_('no source file in %s') % rep, niverr)
         return iret, []
      iret = 0
      nskip = 0
      for f in files:
         tail = os.path.join('...', os.path.basename(os.path.dirname(f)),
                                  os.path.basename(f))
         obj = os.path.basename(f)[:-2]+'.o'

         # ----- skip file if .o more recent than source file
         if not force and os.path.exists(obj) \
               and os.stat(obj).st_ctime > os.stat(f).st_ctime:
            nskip += 1
            if self.verbose or self.debug:
               self.run.VerbStart(_('compiling %s') % tail, verbose=True)
               self.run.VerbIgnore(verbose=True)
         else:
            self.run.VerbStart(_('compiling %s') % tail, verbose=True)
            if self.verbose:
               print
            jret, out = self.run.Shell(cmd+' '+f)
            # ----- avoid to duplicate output
            if not self.verbose:
               self.run.VerbEnd(jret, output=out, verbose=True)
            if jret != 0:
               fmsg.write(out)
               self._mess((_('error during compiling %s (see %s)') % \
                     (f, os.path.abspath(msg))),
                     _('<E>_COMPIL_ERROR'), store=True)
            iret = max(iret, jret)
      self._mess(_('%4d files to compile in %s') % (len(files), rep))
      if nskip > 0:
         self._mess(_('%4d files skipped (object more recent than source)') \
                % nskip)

      fmsg.close()
      os.chdir(prev)
      return iret, files

#-------------------------------------------------------------------------------
   def CompilAster(self, REPREF, repdest='', dbg='nodebug', ignore_ferm=False):
      """Compile all Aster source files, put '.o' files in 'repdest'/DIR
      and DIR_f with DIR=obj or dbg depends on 'dbg'
      (DIR for "real" source files and DIR_f for fermetur).
      """
      self._mess(_('Compilation of source files'), 'TITLE')
      dest_o = {  'nodebug' : self.conf['BINOBJ_NODBG'][0],
                  'debug'   : self.conf['BINOBJ_DBG'][0]    }
      dest_f_o = {'nodebug' : self.conf['BINOBJF_NODBG'][0],
                  'debug'   : self.conf['BINOBJF_DBG'][0]   }

      # ----- define type and destination of '*.o' files for each directory
      if repdest == '':
         repdest = REPREF
      dest = os.path.join(repdest, dest_o[dbg])
      dest_f = os.path.join(repdest, dest_f_o[dbg])
      para = {
         self.conf['SRCC'][0]     : ['C', dest],
         self.conf['SRCFOR'][0]   : ['F', dest],
         self.conf['SRCF90'][0]   : ['F90', dest],
      }
      if not ignore_ferm:
         para[self.conf['SRCFERM'][0]] = ['F', dest_f]

      # ----- check that directories exist
      for obj in [os.path.join(REPREF, rep) for rep in para.keys()]:
         if not os.path.exists(obj):
            self._mess(_('directory does not exist : %s') % obj,
                  _('<E>_FILE_NOT_FOUND'))
      self.run.CheckOK()

      # ----- nobuild list
      nobuild = []
      for d in self.conf['NOBUILD']:
         nobuild.extend([os.path.join(REPREF, d) for d in d.split()])
      # bibf90 dirs automatically added if there is no F90 compiler
      if self.conf['F90'][0] == '' and self.conf['SRCF90'][0] != '':
         nobuild.extend([d for d in \
            glob.glob(os.path.join(REPREF, self.conf['SRCF90'][0], '*'))
            if not d in nobuild])
         self._mess(_("There is no Fortran 90 compiler in 'config.txt'. " \
               "All source files from '%s' were ignored.") % \
               self.conf['SRCF90'][0], _('<A>_IGNORE_F90'), store=True)
      if len(nobuild)>0:
         self._mess(_('These directories will not be built :'))
         for d in nobuild:
            print ' '*10+'>>> %s' % d
      
      # ----- compilation
      iret = 0
      for module in [d for d in para.keys() if d != '']:
         if module != self.conf['SRCFERM'][0]:
            # ----- glob subdirectories
            lrep = glob.glob(os.path.join(REPREF, module, '*'))
         else:
            lrep = (os.path.join(REPREF, module), )
         for rep in lrep:
            if os.path.isdir(rep):
               tail = '...'+re.sub('^'+REPREF, '', rep)
               print
               jret = 0
               if not rep in nobuild:
                  self._mess(_('Compilation from %s directory') % tail)
                  jret, bid = self.Compil(para[module][0], rep, para[module][1], dbg)
               else:
                  self.run.VerbStart(_('Directory %s is in NOBUILD list') % tail, \
                        verbose=True)
                  self.run.VerbIgnore(verbose=True)
               iret = max(iret, jret)

      return iret

#-------------------------------------------------------------------------------
   def Archive(self, repobj, lib):
      """Archive all .o files from 'repobj' into 'lib'
      All args must be local.
      """
      self._mess(_('Archive object files'), 'TITLE')
      prev = os.getcwd()
      # ----- get command line from conf object
      cmd0 = [self.conf['LIB'][0]+' '+lib]

      # ----- check that directories exist
      if not os.path.isdir(repobj):
         self._mess(_('directory does not exist : %s') % repobj,
               _('<E>_FILE_NOT_FOUND'))
      rep = os.path.dirname(lib)
      if not os.path.isdir(rep):
         iret = self.run.MkDir(rep, verbose=True)
      self.run.CheckOK()

      # ----- if force is True, don't use ctime to skip a file
      force = self.run['force']
      if not os.path.exists(lib):
         force = True
         tlib = -1
      else:
         tlib = os.stat(lib).st_ctime

      # ----- work in repobj
      os.chdir(repobj)

      # ----- list of source files
      files = glob.glob('*.o')
      ntot = len(files)
      if ntot == 0:
         self._mess(_('no object file in %s') % rep)
         return 0
      
      # ----- sleep 1 s to be sure than last .o will be older than lib !
      time.sleep(1)

      # ----- use MaxCmdLen to limit command length
      iret = 0
      nskip = 0
      while len(files) > 0:
         cmd = copy.copy(cmd0)
         clen = len(cmd[0])+1
         nadd = 0
         while len(files) > 0 and clen+len(files[0]) < self.run.system.MaxCmdLen-15:
            if not force and tlib > os.stat(files[0]).st_ctime:
               nskip += 1
               bid = files.pop(0)
            else:
               clen = clen+len(files[0])+1
               cmd.append(files.pop(0))
               nadd += 1
         self.run.VerbStart(_('%4d / %4d objects added to archive...') % \
               (nadd, ntot), verbose=True)
         if nadd > 0:
            if self.verbose:
               print
            jret, out = self.run.Shell(' '.join(cmd))
            # ----- avoid to duplicate output
            if not self.verbose:
               self.run.VerbEnd(jret, output=out, verbose=True)
            if jret != 0:
               self._mess(_('error during archiving'), _('<E>_ARCHIVE_ERROR'))
            iret = max(iret, jret)
         else:
            self.run.VerbIgnore(verbose=True)

      self._mess(_('%4d files archived') % ntot)
      if nskip > 0:
         self._mess(_('%4d files skipped (objects older than library)') \
                % nskip)
      os.chdir(prev)
      return iret

#-------------------------------------------------------------------------------
   def Link(self, exe, lobj, libaster, libferm, reptrav):
      """Link a number of object and archive files into an executable.
         exe      : name of the executable to build
         lobj     : list of object files
         libaster : Code_Aster main library
         libferm  : Code_Aster fermetur library
         reptrav  : working directory
      If "python.o" is not in 'lobj' it will be extracted from 'libaster'.
      Other libs are given by ASTER_CONFIG object.
      None argument must not be remote !
      """
      self._mess(_('Build Code_Aster executable'), 'TITLE')
      prev = os.getcwd()
      # ----- check that directories exist
      for obj in lobj+[libaster, libferm]:
         if not os.path.exists(obj):
            self._mess(_('file not found : %s') % obj, _('<E>_FILE_NOT_FOUND'))
      rep = os.path.dirname(exe)
      if not os.path.isdir(rep):
         iret = self.run.MkDir(rep, verbose=True)
      self.run.CheckOK()

      # ----- if force is True, don't use ctime to skip a file
      force = self.run['force']
      if not os.path.exists(exe):
         force = True
      else:
         texe = os.stat(exe).st_ctime
         for f in lobj+[libaster, libferm]:
            if os.stat(f).st_ctime > texe:
               force = True
               self.run.DBG('%s more recent than %s' % (f, exe))
               break

      # ----- work in reptrav
      os.chdir(reptrav)

      # ----- python.o in lobj ?
      mainobj = self.conf['BINOBJ_MAIN'][0]
      if mainobj not in [os.path.basename(o) for o in lobj] and force:
         # get 'ar' command without arguments, if it's not 'ar' &$£#@§!
         cmd = [self.conf['LIB'][0].split()[0]]
         cmd.extend(['-xv', libaster, mainobj])
         cmd = ' '.join(cmd)
         self.run.VerbStart(_('extracting %s from %s...') % \
               (repr(mainobj), os.path.basename(libaster)), verbose=True)
         if self.verbose:
            print
         iret, out = self.run.Shell(cmd)
         # ----- avoid to duplicate output
         if not self.verbose:
            self.run.VerbEnd(iret, output=out, verbose=True)
         if iret != 0:
            self._mess(_('error during extracting %s') % repr(mainobj),
                  _('<F>_LIBRARY_ERROR'))
         lobj.insert(0, mainobj)
      else:
         lobj = [o for o in lobj if os.path.basename(o) == mainobj] +\
                [o for o in lobj if os.path.basename(o) != mainobj]

      # ----- get command line from conf object and args
      cmd = copy.copy(self.conf['LINK'])
      cmd.extend(self.conf['OPTL'])
      cmd.append('-o')
      cmd.append(exe)
      cmd.extend(lobj)
      cmd.append(libaster)
      cmd.extend(self.conf['BIBL'])
      cmd.append(libferm)
      cmd = ' '.join(cmd)

      # ----- run ld or skip this stage
      iret = 0
      tail = os.path.join('...', os.path.basename(rep), os.path.basename(exe))
      self.run.VerbStart(_('creating %s...') % tail, verbose=True)
      if force:
         if self.verbose:
            print
         iret, out = self.run.Shell(cmd)
         # ----- avoid to duplicate output
         if not self.verbose:
            self.run.VerbEnd(iret, output=out, verbose=True)
         if iret != 0:
            self._mess(_('error during linking'), _('<E>_LINK_ERROR'))
      else:
         self.run.VerbIgnore(verbose=True)
         self._mess(_('executable more recent than objects and libs ' \
               'in arguments'))

      os.chdir(prev)
      return iret

#-------------------------------------------------------------------------------
   def PrepEnv(self, REPREF, repdest, dbg='nodebug', lang='', **kargs):
      """Prepare 'repdest' with Code_Aster environment from REPREF and
      given arguments in kargs[k] (k in 'exe', 'cmde', 'ele', 'py', 'unigest').
      Note : - REPREF/elements and kargs['ele'] don't exist when building
         elements for the first time.
             - only PYSUPPR entries from 'unigest' are considered here.
      All elements of kargs can be remote.
      """      
      reptrav = os.path.join(repdest, self.conf['REPPY'][0], '__tmp__')
      prev = os.getcwd()
      # ----- check that files and directories in kargs exist
      for occ in kargs.values():
         obj = occ
         if not type(occ) in (ListType, TupleType):
            obj = [occ]
         for rep in obj:
            if not self.run.Exists(rep):
               self._mess(_('path does not exist : %s') % rep,
                     _('<E>_FILE_NOT_FOUND'))
      if not os.path.isdir(repdest):
         iret = self.run.MkDir(repdest, verbose=True)
      self.run.CheckOK()

      # ----- language
      if lang in ('', 'fr'):
         lang = 'fr'
         suff = ''
      else:
         suff = '_' + lang
      mask_catapy = fmt_catapy % (suff, '*')
      
      # ----- copy REPREF/aster?.exe or kargs['exe']
      iret = 0
      if kargs.has_key('exe'):
         src = kargs['exe']
      else:
         if dbg == 'debug':
            src = os.path.join(REPREF, self.conf['BIN_DBG'][0])
         else:
            src = os.path.join(REPREF, self.conf['BIN_NODBG'][0])
      dest = repdest
      if src != os.path.join(repdest, os.path.basename(src)):
         if not self.run.IsRemote(src) and self.run['symlink'] != False:
            src = self.run.PathOnly(src)
            self.run.VerbStart(_('adding a symbolic link to %s...') % src, verbose=True)
            iret = 0
            output = ''
            try:
               lien = os.path.join(repdest, os.path.basename(src))
               if os.path.exists(lien):
                  self.run.Delete(lien)
               os.symlink(src, lien)
            except OSError, output:
               iret = 4
               self._mess(_('error occurs during creating a symbolic link' \
                           ' from %s to %s') % (src, dest), _('<E>_SYMLINK'))
            self.run.VerbEnd(iret, output, verbose=True)
         else:
            iret = self.run.Copy(dest, src, verbose=True)
      else:
         self.run.VerbStart(_('copying %s...') % src, verbose=True)
         self.run.VerbIgnore(verbose=True)

      # ----- copy REPREF/bibpyt
      self.run.MkDir(os.path.join(repdest, self.conf['REPPY'][0], 'Cata'))
      iret = 0
      src  = os.path.join(REPREF, self.conf['SRCPY'][0], '*')
      dest = os.path.join(repdest, self.conf['REPPY'][0])
      iret = self.run.Copy(dest, src)

      # if 'py' is in kargs
      if kargs.has_key('py'):
         self._mess(_("copy user's python source files"))
         # ----- python source given by user
         lobj = kargs['py']
         if not type(lobj) in (ListType, TupleType):
            lobj = [kargs['py'], ]
         for occ in lobj:
            obj = occ
            if self.run.IsRemote(occ):
               obj = reptrav
               iret = self.run.MkDir(obj)
               if self.run.IsDir(occ):
                  src = os.path.join(occ, '*.py')
               else:
                  src = occ
               iret = self.run.Copy(obj, src, verbose=True)
               files = glob.glob(os.path.join(obj, '*.py'))
            else:
               occ = self.run.PathOnly(occ)
               if self.run.IsDir(occ):
                  files = glob.glob(os.path.join(occ, '*.py'))
               else:
                  files = [occ, ]
            for f in files:
               mod, rep = self.GetCModif('py', f)
               dest = os.path.join(repdest, self.conf['REPPY'][0], rep, mod+'.py')
               iret = self.run.Copy(dest, f, verbose=True)

      # ----- apply PYSUPPR directives from unigest
      if kargs.has_key('unigest'):
         for f in kargs['unigest']['py']:
            self.run.Delete(os.path.join(repdest, \
                  re.sub('^'+self.conf['SRCPY'][0], self.conf['REPPY'][0], f)))

      # ----- copy REPREF/commande or kargs['cmde']
      iret = 0
      if kargs.has_key('cmde'):
         src = kargs['cmde']
      else:
         src = os.path.join(REPREF, self.conf['BINCMDE'][0])
      dest = os.path.join(repdest, self.conf['REPPY'][0], 'Cata')
      if not self.run.IsRemote(src) and self.run['symlink'] != False:
         src = self.run.PathOnly(src)
         # checks catalogue exists
         l_capy = glob.glob(os.path.join(src, mask_catapy))
         if len(l_capy) == 0 and lang != '':
            self._mess(_("no catalogue found for language '%s', " \
                         "use standard (fr) one...") % lang)
            mask_catapy = fmt_catapy % ('', '*')
            l_capy = glob.glob(os.path.join(src, mask_catapy))
         if len(l_capy) == 0:
            self._mess(_('no catalogue found : %s') % os.path.join(src, mask_catapy),
                  _('<F>_FILE_NOT_FOUND'))
         # add symlink
         for f in l_capy:
            root = os.path.splitext(fcapy)[0]
            ext  = os.path.splitext(f)[-1]
            lien = os.path.join(dest, root + ext)
            self.run.VerbStart(_('adding a symbolic link %s to %s...') % \
                                (os.path.basename(lien), f), verbose=True)
            iret = 0
            output = ''
            try:
               if os.path.exists(lien):
                  self.run.Delete(lien)
               os.symlink(f, lien)
            except OSError, output:
               iret = 4
               self._mess(_('error occurs during creating a symbolic link' \
                            ' from %s to %s') % (src, dest), _('<E>_SYMLINK'))
            self.run.VerbEnd(iret, output, verbose=True)
      else:
         iret = self.run.Copy(dest, os.path.join(src, mask_catapy), verbose=True)
         for f in glob.glob(os.path.join(dest, mask_catapy)):
            root = os.path.splitext(fcapy)[0]
            ext  = os.path.splitext(f)[-1]
            os.rename(f, os.path.join(dest, root + ext))

      # ----- copy REPREF/elements or kargs['ele']
      iret = 0
      if kargs.has_key('ele'):
         src = kargs['ele']
      else:
         src = os.path.join(REPREF, self.conf['BINELE'][0])
      if self.run.Exists(src) and src != os.path.join(repdest, 'elem.1'):
         # symlink not allowed for elem.1 (rw)
         iret = self.run.Copy(os.path.join(repdest, 'elem.1'), src, verbose=True)
      else:
         self.run.VerbStart(_('copying %s...') % src, verbose=True)
         self.run.VerbIgnore(verbose=True)

      # ----- result directories
      os.chdir(repdest)
      self.run.MkDir('RESU_ENSIGHT')
      self.run.MkDir('REPE_OUT')

      os.chdir(prev)
      self.run.Delete(reptrav)
      return iret

#-------------------------------------------------------------------------------
   def CompilCapy(self, REPREF, reptrav, exe='', cmde='', i18n=False, **kargs):
      """Compile commands catalogue from REPREF/commande using 'exe'
      and puts result in 'cmde'. kargs can contain :
         capy    : list of user's capy files
         unigest : GetUnigest dict with "CATSUPPR catapy module" entries
      All args can be remote.
      If `i18n` is True, translates cata.py.
      """
      self._mess(_('Compilation of commands catalogue'), 'TITLE')
      prev = os.getcwd()
      if exe == '':
         exe = os.path.join(REPREF, self.conf['BIN_NODBG'][0])
      if cmde == '':
         cmde = os.path.join(REPREF, self.conf['BINCMDE'][0])
      # ----- check that files and directories in kargs exist
      for occ in kargs.values():
         obj = occ
         if not type(occ) in (ListType, TupleType):
            obj = [occ]
         for rep in obj:
            if not self.run.Exists(rep):
               self._mess(_('path does not exist : %s') % rep,
                     _('<E>_FILE_NOT_FOUND'))
      if not os.path.isdir(cmde):
         iret = self.run.MkDir(cmde, verbose=True)
      if not os.path.isdir(reptrav):
         iret = self.run.MkDir(reptrav, verbose=True)
      self.run.CheckOK()

      # ----- work in reptrav
      os.chdir(reptrav)

      # ----- if force is True, don't use ctime to skip a file
      force = self.run['force']

      # ----- copy capy from REPREF
      if os.path.exists(self.conf['SRCCAPY'][0]):
         self.run.Delete(self.conf['SRCCAPY'][0])
      iret = self.run.Copy(self.conf['SRCCAPY'][0],
            os.path.join(REPREF, self.conf['SRCCAPY'][0]), verbose=True)

      # if 'capy' is in kargs
      if kargs.has_key('capy'):
         force = True
         self._mess(_("copy user's catalogues"))
         # ----- catapy given by user
         lobj = kargs['capy']
         if not type(obj) in (ListType, TupleType):
            lobj = [kargs['capy'], ]
         for occ in lobj:
            obj = occ
            if self.run.IsRemote(occ):
               obj = os.path.join(self.conf['SRCCAPY'][0],
                                  '__tmp__'+os.path.basename(occ))
               iret = self.run.MkDir(obj)
               if self.run.IsDir(occ):
                  src = os.path.join(occ, '*.capy')
               else:
                  src = occ
               iret = self.run.Copy(obj, src)
               files = glob.glob(os.path.join(obj, '*.capy'))
            else:
               occ = self.run.PathOnly(occ)
               if self.run.IsDir(occ):
                  files = glob.glob(os.path.join(occ, '*.capy'))
               else:
                  files = [occ, ]
            for f in files:
               mod, rep = self.GetCModif('capy', f)
               dest = os.path.join(self.conf['SRCCAPY'][0], rep)
               iret = self.run.Copy(dest, f, verbose=True)
               if not rep in self.repcatapy:
                  self._mess(_('unknown module name : %s') % rep, _('<A>_ALARM'))

      # ----- apply CATSUPPR directives from unigest
      if kargs.has_key('unigest'):
         force = True
         for f in kargs['unigest']['capy']:
            self.run.Delete(os.path.join(reptrav, f))

      # ----- build cata.py
      if os.path.exists(fcapy):
         self.run.Delete(fcapy)

      # ----- test if compilation can be skipped
      if force or not os.path.exists(os.path.join(cmde, fcapy)):
         force = True
      else:
         tcpy = os.stat(os.path.join(cmde, fcapy)).st_ctime
         ltest = glob.glob(os.path.join(self.conf['SRCCAPY'][0], '*', '*.capy'))
         for f in ltest:
            if os.stat(f).st_ctime >= tcpy:
               force = True
               break

      if not force:
         self.run.VerbStart(_('compilation of commands'), verbose=True)
         self.run.VerbIgnore(verbose=True)
      else:

         fo = open(fcapy, 'w')
         for rep in self.repcatapy:
            for capy in glob.glob(os.path.join(self.conf['SRCCAPY'][0], rep, '*.capy')):
               fo2 = open(capy, 'r')
               fo.write(fo2.read())
               fo2.close()
         fo.close()

         # ----- compile cata.py
         dtmp = kargs.copy()
         dtmp['exe'] = exe
         dtmp['cmde'] = reptrav
         if kargs.has_key('py'):
            dtmp['py'] = kargs['py']
         self.PrepEnv(REPREF, reptrav, **dtmp)
         
         self.run.VerbStart(_('compilation of commands'), verbose=True)
         cmd_import = """
import sys
sys.path.insert(0, "%s")
iret = 0
try:
   from Cata import cata
   from Cata.cata import JdC
   cr = JdC.report()
   if not cr.estvide() :
      iret = 4
      print ">> Catalogue de commandes : DEBUT RAPPORT"
      print cr
      print ">> Catalogue de commandes : FIN RAPPORT"
except:
   iret = 4
   import traceback
   traceback.print_exc(file=sys.stdout)
sys.exit(iret)
""" % self.conf['REPPY'][0]
         cmd_import_file = os.path.join(reptrav, 'cmd_import.py')
         open(cmd_import_file, 'w').write(cmd_import)
         
         if self.verbose:
            print
         cmd = [os.path.join('.', os.path.basename(exe))]
         cmd.append(cmd_import_file)
         iret, out = self.run.Shell(' '.join(cmd))
         # ----- avoid to duplicate output
         if not self.verbose:
            self.run.VerbEnd(iret, output=out, verbose=True)
         
         if iret != 0:
            self._mess(_('error during compiling %s') % fcapy,
                  _('<E>_CATAPY_ERROR'))
         else:
            kret = self.run.Copy(cmde,
                   os.path.join(self.conf['REPPY'][0], 'Cata', fcapy_), verbose=True)
            self._mess(_('Commands catalogue successfully compiled'), 'OK')

         
         # --- translations...
         if i18n:
            pattern = '#:LANG:%s:LANG_END:'
            old     = pattern % '.*'
            exp     = re.compile(old, re.MULTILINE | re.DOTALL)
            for lang in [l for l in self.conf['I18N'] if not l in ('', 'fr')]:
               self.run.VerbStart(_('working for i18n (language %s)...') % lang, verbose=True)
               print
               new  = pattern % (os.linesep + "lang = '%s'" % lang + os.linesep + '#')
               cata_out  = fmt_catapy % ('_' + lang, '')
               cata_out_ = fmt_catapy % ('_' + lang, '*')
               tmpl_dict = os.path.join(self.run['rep_trav'],
                                        'template_dict_%s.py' % lang)
               # fake aster module
               open(os.path.join(self.conf['REPPY'][0], 'aster.py'), 'w').write("""
# fake aster module
""")
               sys.path.insert(0, self.conf['REPPY'][0])
               iret = 0
               try:
                  from Cata.i18n_cata import i18n_make_cata_fich
                  txt_cata = i18n_make_cata_fich(fichier_dtrans=tmpl_dict, lang=lang)
                  txt_orig = []
                  fo = open(cata_out, 'w')
                  # prendre l'entete
                  for rep in self.repcatapy[:1]:
                     for capy in glob.glob(os.path.join(self.conf['SRCCAPY'][0], rep, '*.capy')):
                        txt0 = open(capy, 'r').read()
                        txt2 = exp.sub(new, txt0)
                        fo.write(txt2)
                  # reprendre les imports
                  for rep in self.repcatapy[1:]:
                     for capy in glob.glob(os.path.join(self.conf['SRCCAPY'][0], rep, '*.capy')):
                        fo.write(get_txt_capy(capy))
                        txt_orig.append(open(capy, 'r').read())
                  fo.write(os.linesep)
                  fo.write(txt_cata)
                  fo.write(os.linesep)
                  fo.write(os.linesep.join(txt_orig))
                  fo.close()
               except:
                  iret = 4
                  import traceback
                  traceback.print_exc(file=sys.stdout)
                        
               # ----- avoid to duplicate output
               if not self.verbose:
                  self.run.VerbEnd(iret, output=out, verbose=True)
               
               if iret != 0:
                  self._mess(_("error during building %s") % cata_out, _('<E>_I18N_ERROR'))
               else:
                  kret = self.run.Copy(cmde, cata_out_, verbose=True)
                  self._mess(_('Commands catalogue successfully translated (language %s)') % lang, 'OK')

      os.chdir(prev)
      self.run.Delete(reptrav)
      return iret

#-------------------------------------------------------------------------------
   def CompilEle(self, REPREF, reptrav, ele='', **kargs):
      """Compile elements catalogue as 'ele' from cata_ele.pickled.
      kargs must contain arguments for a Code_Aster execution (i.e. for PrepEnv)
      and optionnaly :
         cata    : a list of user's elements files
         unigest : GetUnigest dict with "CATSUPPR catalo module" entries
         (make_surch_offi needs unigest filename)
      All args can be remote.
      """
      self._mess(_('Compilation of elements catalogue'), 'TITLE')
      required = ('exe', 'cmde')
      if ele == '':
         ele = os.path.join(REPREF, self.conf['BINELE'][0])
      prev = os.getcwd()
      # ----- check for required arguments
      for a in required:
         if not kargs.has_key(a):
            self._mess('(CompilEle) '+_('argument %s is required') % a,
                  _('<E>_PROGRAM_ERROR'))
      
      # ----- check that files and directories in kargs exist
      for occ in kargs.values():
         obj = occ
         if not type(occ) in (ListType, TupleType):
            obj = [occ]
         for rep in obj:
            if not self.run.Exists(rep):
               self._mess(_('path does not exist : %s') % rep,
                     _('<E>_FILE_NOT_FOUND'))
      if not os.path.isdir(reptrav):
         iret = self.run.MkDir(reptrav, verbose=True)
      self.run.CheckOK()

      # ----- work in repobj
      os.chdir(reptrav)

      # ----- if force is True, don't use ctime to skip a file
      force = self.run['force']

      # 1. ----- if 'cata' is in kargs
      if kargs.has_key('cata'):
         force = True
         self._mess(_("copy user's catalogues"))
         fo = open('surch.cata', 'w')
         # ----- catapy given by user
         lobj = kargs['cata']
         if not type(obj) in (ListType, TupleType):
            lobj = [kargs['cata'], ]
         for occ in lobj:
            obj = occ
            if self.run.IsRemote(occ):
               obj = os.path.join(reptrav, '__tmp__'+os.path.basename(occ))
               iret = self.run.MkDir(obj)
               if self.run.IsDir(occ):
                  src = os.path.join(occ, '*.cata')
               else:
                  src = occ
               iret = self.run.Copy(obj, src)
               files = glob.glob(os.path.join(obj, '*.cata'))
            else:
               occ = self.run.PathOnly(occ)
               if self.run.IsDir(occ):
                  files = glob.glob(os.path.join(occ, '*.cata'))
               else:
                  files = [occ, ]
            for f in files:
               fo2 = open(f, 'r')
               self.run.VerbStart(_('adding %s') % f, verbose=True)
               fo.write(fo2.read())
               self.run.VerbEnd(iret=0, output='', verbose=True)
               fo2.close()
         fo.close()

      # 2. ----- compile elements with MAKE_SURCH_OFFI
      self.PrepEnv(REPREF, reptrav, **kargs)
      iret = 0
      cmd = [os.path.join('.', os.path.basename(kargs['exe']))]
      cmd.append(os.path.join(self.conf['REPPY'][0], self.conf['MAKE_SURCH_OFFI'][0]))
      cmd.extend(self.conf['REPPY'])
      cmd.append('MAKE_SURCH')
      cmd.append('surch.cata')
      funig = 'unigest_bidon'
      if kargs.has_key('unigest'):
         force = True
         funig = kargs['unigest']['filename']
         if self.run.IsRemote(funig):
            funig = 'unigest'
            kret = self.run.Copy(funig, kargs['unigest']['filename'])
         else:
            funig = self.run.PathOnly(funig)
      cmd.append(funig)
      cmd.append(os.path.join(REPREF, self.conf['BINPICKLED'][0]))
      cmd.append('fort.4')
      self.run.VerbStart(_('pre-compilation of elements with %s') % \
            os.path.basename(self.conf['MAKE_SURCH_OFFI'][0]), verbose=True)

      # 2.1. ----- test if compilation can be skipped
      if force or not os.path.exists(ele):
         force = True
      else:
         tele = os.stat(ele).st_ctime
         ltest = [os.path.join(REPREF, self.conf['BINPICKLED'][0]), ]
         ltest.extend(glob.glob(os.path.join(self.conf['REPPY'][0], '*', '*.py')))
         for f in ltest:
            if os.stat(f).st_ctime >= tele:
               force = True
               self.run.DBG('%s more recent than %s' % (f, ele))
               break

      if not force:
         self.run.VerbIgnore(verbose=True)
      else:
         if self.verbose:
            print
         jret, out = self.run.Shell(' '.join(cmd))
         # ----- avoid to duplicate output
         if not self.verbose:
            self.run.VerbEnd(jret, output=out, verbose=True)
         if jret != 0:
            self._mess(_('error during pre-compilation of elements'),
                  _('<F>_CATAELE_ERROR'))

      # 3. ----- build elem.1 with kargs['exe']
      cmd = [os.path.join('.', os.path.basename(kargs['exe']))]
      cmd.append(os.path.join(self.conf['REPPY'][0], self.conf['ARGPYT'][0]))
      cmd.extend(self.conf['ARGEXE'])
      cmd.append('-commandes fort.1 -mem 8 -rep none -tpmax 60')
      fo = open('fort.1', 'w')
      fo.write("""
# COMPILATION DU CATALOGUE D'ELEMENT
DEBUT ( CATALOGUE = _F( FICHIER = 'CATAELEM' , UNITE = 4 ))
MAJ_CATA ( ELEMENT = _F())
FIN()
""")
      fo.close()
      self.run.VerbStart(_('compilation of elements with %s') \
            % kargs['exe'], verbose=True)
      
      if not force:
         self.run.VerbIgnore(verbose=True)
      else:
         if self.verbose:
            print
         jret, out = self.run.Shell(' '.join(cmd))
         # ----- diagnostic of Code_Aster execution
         diag = self.GetDiag()[0]
         if self.run.GetGrav(diag) < self.run.GetGrav('<S>') \
             and os.path.exists('elem.1'):
            jret = 0
         else:
            jret = 4
         # ----- avoid to duplicate output
         if not self.verbose:
            self.run.VerbEnd(jret, output=out, verbose=True)

         if jret == 0:
            self._mess(_('Elements catalogue successfully compiled'), diag)
         else:
            self._mess(_('error during compilation of elements'),
                  _('<F>_CATAELE_ERROR'))

         # 4. ----- copy elements to 'ele'
         iret = self.run.Copy(ele, 'elem.1', verbose=True)

      os.chdir(prev)
      self.run.Delete(reptrav)
      return iret

#-------------------------------------------------------------------------------
   def MakePickled(self, REPREF, reptrav, repdest='', **kargs):
      """Make 'repdest'/cata_ele.pickled.
      kargs can contain :
         exe     : Code_Aster executable or Python interpreter (can be remote)
      """
      self._mess(_('Prepare cata_ele.pickled'), 'TITLE')
      if repdest == '':
         repdest = REPREF
      prev = os.getcwd()
      # ----- check if reptrav exists
      if not os.path.isdir(reptrav):
         iret = self.run.MkDir(reptrav, verbose=True)

      # ----- work in repobj
      os.chdir(reptrav)

      # ----- if force is True, don't use ctime to skip a file
      force = self.run['force']

      # ----- use REPREF/aster?.exe or kargs['exe']
      iret = 0
      if kargs.has_key('exe'):
         exe = kargs['exe']
         if self.run.IsRemote(exe):
            kret = self.run.Copy(reptrav, exe)
            exe = os.path.join(reptrav, os.path.basename(exe))
         else:
            exe = self.run.PathOnly(exe)
      else:
         exe = os.path.join(REPREF, self.conf['BIN_NODBG'][0])
      
      # ----- call MAKE_CAPY_OFFI
      cmd = [exe, ]
      cmd.append(os.path.join(REPREF, self.conf['SRCPY'][0], self.conf['MAKE_CAPY_OFFI'][0]))
      cmd.append(os.path.join(REPREF, self.conf['SRCPY'][0]))
      cmd.append('TRAV_PICKLED')
      cmd.append(os.path.join(REPREF, self.conf['SRCCATA'][0]))
      cata_pickled = os.path.basename(self.conf['BINPICKLED'][0])
      cmd.append(cata_pickled)

      # ----- test if compilation can be skipped
      if not os.path.exists(os.path.join(repdest, self.conf['BINPICKLED'][0])):
         force = True
      else:
         tpick = os.stat(os.path.join(repdest, self.conf['BINPICKLED'][0])).st_ctime
         ltest = glob.glob(os.path.join(REPREF, self.conf['SRCCATA'][0], '*', '*.cata'))
         ltest.extend(glob.glob(os.path.join(REPREF, self.conf['SRCPY'][0], '*', '*.py')))
         for f in ltest:
            if os.stat(f).st_ctime >= tpick:
               force = True
               break

      self.run.VerbStart(_('build cata_ele.pickled with %s') % \
            os.path.basename(self.conf['MAKE_CAPY_OFFI'][0]), verbose=True)
      
      if not force:
         self.run.VerbIgnore(verbose=True)
      else:
         if self.verbose:
            print
         jret, out = self.run.Shell(' '.join(cmd))
         # ----- avoid to duplicate output
         if not self.verbose:
            self.run.VerbEnd(jret, output=out, verbose=True)
         if jret == 0 and os.path.exists(cata_pickled):
            self._mess(_('cata_ele.pickled successfully created'), 'OK')
         else:
            self._mess(_('error during making cata_ele.pickled'),
                  _('<F>_PICKLED_ERROR'))

         # ----- copy cata_ele.pickled to 'repdest'
         iret = self.run.Copy(os.path.join(repdest, self.conf['BINPICKLED'][0]),
                              cata_pickled, verbose=True)

      os.chdir(prev)
      self.run.Delete(reptrav)
      return iret

#-------------------------------------------------------------------------------
   def GetUnigest(self, funig):
      """Build a dict of the file names by parsing unigest file 'funig' where
      the keys are 'f', 'f90', 'c', 'py', 'capy', 'cata', 'test' and 'fdepl'.
      funig can be remote.
      WARNING : 'f90' not yet supported !
      """
      dico = {}
      for cat in ('f', 'f90', 'c', 'py', 'capy', 'cata', 'test', 'fdepl'):
         dico[cat] = []

      # ----- keep a reference to the file for CompilEle
      dico['filename'] = funig

      # ----- check that file exists
      if not self.run.Exists(funig):
         self._mess(_('file not found : %s') % funig, _('<A>_ALARM'))
         return dico

      # ----- copy funig if it's a remote file
      if self.run.IsRemote(funig):
         name = '__tmp__.unigest'
         kret = self.run.Copy(name, funig)
      else:
         name = self.run.PathOnly(funig)

      # parse unigest
      fo = open(name, 'r')
      for line in fo:
         mat = re.search('^ *FORSUPPR +([-_a-zA-Z0-9]+) +([-_a-zA-Z0-9]+)', line)
         if mat:
            nam = mat.group(1).lower()+'.f'
            mod = mat.group(2).lower()
            dico['f'].append(os.path.join(self.conf['SRCFOR'][0], mod, nam))
         mat = re.search('^ *F90SUPPR +([-_a-zA-Z0-9]+) +([-_a-zA-Z0-9]+)', line)
         if mat:
            nam = mat.group(1).lower()+'.f'
            mod = mat.group(2).lower()
            dico['f'].append(os.path.join(self.conf['SRCF90'][0], mod, nam))
         mat = re.search('^ *CSUPPR +([-_a-zA-Z0-9]+) +([-_a-zA-Z0-9]+)',
               line)
         if mat:
            nam = mat.group(1).lower()+'.c'
            mod = mat.group(2).lower()
            dico['c'].append(os.path.join(self.conf['SRCC'][0], mod, nam))
         mat = re.search('^ *PYSUPPR +([-_a-zA-Z0-9]+) +([-_a-zA-Z0-9]+)',
               line)
         if mat:
            nam = mat.group(1)+'.py'
            mod = mat.group(2)
            dico['py'].append(os.path.join(self.conf['SRCPY'][0], mod, nam))
         mat = re.search('^ *CATSUPPR +([-_a-zA-Z0-9]+) +([-_a-zA-Z0-9]+)', line)
         if mat:
            nam = mat.group(1).lower()
            mod = mat.group(2).lower()
            if mod in self.repcatapy:
               dico['capy'].append(os.path.join(self.conf['SRCCAPY'][0], mod, nam+'.capy'))
            elif mod in self.repcatalo:
               dico['cata'].append(os.path.join(self.conf['SRCCATA'][0], mod, nam+'.cata'))
            else:
               self._mess(_('unknown type %s in unigest file') % mod,
                     _('<A>_ALARM'))
         mat = re.search('^ *TESSUPPR +([-_a-zA-Z0-9]+)', line)
         if mat:
            nam = mat.group(1)+'.*'
            dico['test'].append(os.path.join(self.conf['SRCTEST'][0], nam))
         mat = re.search('^ *FORDEPLA +([-_a-zA-Z0-9]+) +([-_a-zA-Z0-9]+)' \
               ' +([-_a-zA-Z0-9]+)', line)
         if mat:
            nam = mat.group(1).lower()+'.f'
            old = mat.group(2).lower()
            new = mat.group(3).lower()
            dico['fdepl'].append([os.path.join(self.conf['SRCFOR'][0], old, nam),
                  os.path.join(self.conf['SRCFOR'][0], new, nam)])
      fo.close()
      return dico

#-------------------------------------------------------------------------------
   def GetDiag(self, err='fort.9', resu='fort.8', mess='fort.6', cas_test=False):
      """Return the diagnostic after a Code_Aster execution
      as a list : [diag, tcpu, tsys, ttotal]
         err   : error file
         resu  : result file (None = ignored)
         mess  : messages file (None = ignored)
      """
      diag, tcpu, tsys, ttot = _('<F>_ABNORMAL_ABORT'), 0., 0., 0.
      
      # 1. ----- parse error file
      diag2, stop = _diag_erre(err)
      diag = diag2 or diag
      if stop:
         self.run.ASSERT(diag != None)
      else:
         # if .erre doesn't exist, same analysis with .resu
         if diag is None:
            diag, stop = _diag_erre(resu)
      
         if stop:
            self.run.ASSERT(diag != None)
         else:
            # 2. ----- parse result file
            diag, tcpu, tsys, ttot = _diag_resu(resu, cas_test)
            # if resu doesn't exist, trying with .mess
            if diag is None:
               diag = _diag_mess(mess) or _('NO_RESU_FILE')
      
      return diag, tcpu, tsys, ttot

#-------------------------------------------------------------------------------
   def GetCModif(self, typ, fich):
      """
         Retourne, pour `typ` = 'capy'/'py', le nom du module et du package.
         `fich` est local.
      """
      para = {
         'capy'   : { 'comment' : '#&', 'nmin' : 3 },
         'py'     : { 'comment' : '#@', 'nmin' : 4 },
      }
      if typ not in para.keys():
         self._mess(_('invalid type : %s') % typ, _('<F>_PROGRAM_ERROR'))
      
      lig = open(fich, 'r').readline().split()
      if len(lig) < para[typ]['nmin'] or \
        (len(lig) > 0 and lig[0] != para[typ]['comment']):
         self._mess(_("invalid first line of file : %s\n" \
                      "      Should start with '%s' and have at least %s fields") \
            % (os.path.basename(fich), para[typ]['comment'], para[typ]['nmin']),
            _('<F>_INVALID_FILE'))
      
      if typ == 'py':
         mod = lig[2]
         rep = lig[3]
      elif typ == 'capy':
         mod = os.path.splitext(os.path.basename(fich))[0]
         rep = lig[2].lower()
      
      return mod, rep


#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
def get_txt_capy(capy):
   """
      On utilise le fait qu'il n'y a rien après la définition de l'OPER, PROC ou MACRO.
   """
   txt = open(capy, 'r').read()
   expr = re.compile('^[A-Za-z_0-9]+ *= *[OPERPROCMACROFORM]+ *\(.*',
                     re.MULTILINE | re.DOTALL)
   sans_op = expr.sub('', txt)
   keep = ['']
   for line in sans_op.splitlines():
      if not re.search('^ *#', line):
         keep.append(line)
   keep.append('')
   return os.linesep.join(keep)

#-------------------------------------------------------------------------------
def re_search(content, **kargs):
   """Search the regular expression `pattern` in content.
   If `string` is provided, it is escaped and `pattern` must contain "%s".
   """
   flag = kargs.get('flag', re.MULTILINE)
   what = kargs.get('result', 'bool')
   pattern = kargs.get('pattern', '%s')
   
   if kargs.get('string'):
      pattern = pattern % re.escape(kargs['string'])
   
   expr = re.compile(pattern, flag)
   
   if what == 'bool':
      res = expr.search(content) != None
   elif what == 'number':
      res = len(expr.findall(content))
   elif what == 'value':
      res = expr.findall(content)
   else:
      res = None
   return res

#-------------------------------------------------------------------------------
def _diag_erre(err):
   """
      Diagnostic à partir du fichier '.erre'.
   """
   diag = None
   stop = False
   if err is None or not os.path.exists(err):
      pass
   else:
      txt = open(err, 'r').read()
      
      n_debut = re_search(txt, string="-- CODE_ASTER -- VERSION", result='number')
      n_fin   = re_search(txt,
            string="""<I> <FIN> ARRET NORMAL DANS "FIN" PAR APPEL A "JEFINI".""",
            result='number')
      
      alarm = re_search(txt, pattern='^ *. *%s', string='<A>')
      fatal = re_search(txt, pattern='^ *. *%s', string='<F>')
      
      if fatal:
         expir = re_search(txt, pattern='^ *. *%s',
               string='<F> <INIAST> VERSION EXPIREE')
         superv = re_search(txt, pattern='^ *. *%s',
               string='<F> <SUPERVISEUR> ARRET SUR ERREUR(S) UTILISATEUR')
      else:
         expir  = False
         superv = False
      
      errs = re_search(txt, pattern='^ *. *%s', string='<S>')
      
      if errs:
         mem = re_search(txt, string='MEMOIRE INSUFFISANTE POUR ALLOUER')
         arretcpu = re_search(txt, pattern='^ *%s',
               string='ARRET PAR MANQUE DE TEMPS CPU')
      
      ok = n_debut == n_fin and n_fin > 0
      if ok:
         diag = 'OK'
         if alarm:
            diag = _('<A>_ALARM')
      # fatal error
      if fatal:
         diag = _('<F>_ERROR')
         if superv:
            diag = _('<F>_SUPERVISOR')
         elif expir:
            diag = _('<F>_EXPIRED_VERS')
      # "S" (recoverable) error
      if errs:
         diag = _('<S>_ERROR')
         if arretcpu:
            diag = _('<S>_CPU_LIMIT')
         elif mem:
            diag = _('<S>_MEMORY_ERROR')
      if not ok or fatal or errs:
         stop = True
   
   return diag, stop

#-------------------------------------------------------------------------------
def _diag_resu(resu, cas_test):
   """
      Diagnostic à partir du fichier '.resu'.
   """
   diag, tcpu, tsys, ttot = None, 0., 0., 0.
   # 2. ----- parse result file
   if resu is None or not os.path.exists(resu):
      pass
   else:
      # OK / NOOK
      ok = False
      nook = False
      # <A>_ALARME
      alarm = False
      diag0 = diag
      
      txt = open(resu, 'r').read()
      ok = re_search(txt, pattern='^ *OK +')
      nook = re_search(txt, pattern='^ *NOOK +')
      alarm = re_search(txt, pattern='^ *. *%s', string='<A>')
      
      value = re_search(txt, result='value',
            pattern='TOTAL_JOB +: +([0-9\.]+) +: +([0-9\.]+) +: +([0-9\.]+)')
      if len(value) > 0:
         tcpu, tsys, ttot = value[0]
      
      if nook:
         diag0 = 'NOOK_TEST_RESU'
      else:
         if cas_test and not ok:
            diag0 = 'NO_TEST_RESU'
         else:
            diag0 = 'OK'
            if alarm:
               diag0 = _('<A>_ALARM')
      diag = diag0
   
   return diag, tcpu, tsys, ttot

#-------------------------------------------------------------------------------
def _diag_mess(mess):
   """
      Essaie de trouver des pistes supplémentaires si le .mess existe.
   """
   diag = None
   if mess is None or not os.path.isfile(mess):
      pass
   else:
      txt = open(mess, 'r').read()
      syntax = re_search(txt, string="ERREUR A LA VERIFICATION SYNTAXIQUE") or \
               re_search(txt, pattern="ERREUR .*ACCAS")
      if syntax:
         diag = '<F>_SYNTAX_ERROR'
   
   return diag




# -*- coding: iso-8859-1 -*-

# ==============================================================================
# COPYRIGHT (C) 1991 - 2003  EDF R&D                  WWW.CODE-ASTER.ORG
# THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
# IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
# THE FREE SOFTWARE FOUNDATION; EITHER VERSION 2 OF THE LICENSE, OR
# (AT YOUR OPTION) ANY LATER VERSION.
#
# THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
# WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. SEE THE GNU
# GENERAL PUBLIC LICENSE FOR MORE DETAILS.
#
# YOU SHOULD HAVE RECEIVED A COPY OF THE GNU GENERAL PUBLIC LICENSE
# ALONG WITH THIS PROGRAM; IF NOT, WRITE TO EDF R&D CODE_ASTER,
#    1 AVENUE DU GENERAL DE GAULLE, 92141 CLAMART CEDEX, FRANCE.
# ==============================================================================

"""
   Some utilities
"""

__revision__ = "$Id: utils.py 2579 2006-12-11 08:47:28Z courtois $"

__all__ = ['get_absolute_path', 'get_absolute_dirname',
           'search_cmd', 'change_text_file', 'IterVersion']

# ----- differ messages translation
def _(mesg):
   return mesg

import os
import re

#-------------------------------------------------------------------------------
def get_absolute_path(path):
   """Retourne le chemin absolu en suivant les liens éventuels.
   """
   if os.path.islink(path):
      path = os.path.realpath(path)
   res = os.path.normpath(os.path.abspath(path))
   return res

def get_absolute_dirname(path):
   """Retourne le chemin absolu en suivant les liens éventuels.
   """
   res = os.path.normpath(os.path.join(get_absolute_path(path), os.pardir))
   return res

#-------------------------------------------------------------------------------
def search_cmd(pattern, fich, delimiter=' ;='):
   """Retourne la commande contenant `pattern` dans `fich`.
   """
   cmd = None
   content = open(fich, 'r').read()
   sep = '[%s]*' % re.escape(delimiter)
   expr = re.compile('%(sep)s(.*?%(pattern)s.*?)%(sep)s' % {
      'pattern' : pattern,
      'sep'     : sep,
   }, re.MULTILINE)
   mat = expr.search(content)
   if mat != None:
      cmd = mat.group(1)
      for char in delimiter:
         lsplit = cmd.split(char)
         cmd = ' '.join([s for s in lsplit if re.search(pattern, s)])
   return cmd

#-------------------------------------------------------------------------------
def _chgline(line, dtrans, delimiter):
   """Change all strings by their new value provided by 'dtrans' dictionnary
   in the string 'line'.
   """
   delim = re.escape(delimiter)
   for old, new in dtrans.items():
      line = re.sub(delim + old + delim, str(new), line)
   return line

def change_text_file(filename, dtrans, delimiter, keep=False, ext='.orig'):
   """Change all strings by their new value provided by 'dtrans' dictionnary
   in the existing file 'filename'.
   """
   iret=0
   fold = filename + ext
   try:
      os.rename(filename, fold)
   except OSError:
      return 4
   fsrc=open(fold, 'r')
   fnew=open(filename, 'w')
   for line in fsrc:
      fnew.write(_chgline(line, dtrans, delimiter))
   fnew.close()
   fsrc.close()
   if not keep:
      os.remove(fold)
   return iret

#-------------------------------------------------------------------------------
class IterVersionError:
   pass

class IterVersion:
   """Classe pour itérer sur les sous-versions d'un produit.
   """
   def __init__(self, vini, *delta):
      """Version de la forme n1.n2.n3, et on va jusqu'à l'indice
          "n1+delta1"."n2+delta2"."n3+delta3".
      """
      self.l_vini = self.v2l(vini)
      self.l_vcur = self.l_vini[:]
      self.delta  = delta[:]
      self.imax   = len(self.delta) - 1
      if len(self.l_vini) != len(self.delta):
         raise IterVersionError, 'number of delta does not match version'

   def v2l(self, vers):
      """'1.4.7' => [1, 4, 7]
      """
      try:
         liste = [int(ind) for ind in vers.split('.')]
      except Exception, msg:
         raise IterVersionError, msg
      return liste

   def l2v(self, liste):
      """[1, 4, 7] => '1.4.7'
      """
      return '.'.join(['%s' % val for val in liste])

   def possible(self):
      """Listes des valeurs suivantes possibles
      """
      self.l_poss = [range(self.l_vcur[i] + 1, self.l_vini[i] + self.delta[i] + 1) \
               for i in range(self.imax + 1)]

   def next(self):
      """Retourne la version suivant la version courante selon 'stage'.
      """
      ok = False
      i = self.imax
      self.possible()
      i = self.imax
      while i >= 0 and not ok:
         if len(self.l_poss[i]) > 0:
            self.l_vcur[i] = self.l_poss[i].pop(0)
            ok = True
         else:
            self.possible()
            self.l_vcur[i] = 0
            self.l_vini[i] = 0
            i -= 1
      if not ok:
         raise StopIteration
      return self.l2v(self.l_vcur)

   def jump(self):
      """Saute à la sous-version suivante.
      """
      self.l_vini[self.imax] = -9999
      
   def __iter__(self):
      """Returns the iterator itself.
      """
      return self

#-------------------------------------------------------------------------------
if __name__ == '__main__':
   itv = IterVersion('1.4.7', 1, 4, 5)
   for i, vers in enumerate(itv):
      if i > 6:
         itv.jump()
      print i, vers

